/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2003 - 2007 Funambol, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 *
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */


package com.funambol.mailclient.ui.view;

import com.funambol.mail.Address;
import com.funambol.mail.MailException;
import com.funambol.mailclient.cm.Contact;
import com.funambol.mailclient.loc.Localization;
import com.funambol.mailclient.ui.controller.UIController;
import com.funambol.mailclient.ui.utils.UiUtils;
import com.funambol.util.Log;
import com.funambol.util.StringUtil;
import javax.microedition.lcdui.Display;
import javax.microedition.lcdui.Font;
import javax.microedition.lcdui.Graphics;



/**
 * a class that describes a contactitem objec.
 * this class provides methods for handling the state of the contact
 * (unselected, selected as TO, CC, BCC), for managing the selected email address
 * (default, secondary or tertiary address), for creating an Address from the
 * ContactItem and for painting the contactitem on a given graphics object
 */
public class ContactItem {
    /**
     * ADDRESS_STATE_NONE is composed by Address.TO|Address.CC|Address.BCC and shifted.
     * this way we're sure that ADDRESS_STATE_NONE is different from each of Address.TO,
     * Address.CC, Address.BCC
     */
    public static final int ADDRESS_STATE_NONE=( Address.TO | Address.CC | Address.BCC ) <<1;


    /**
     * current state of the contact. one of ContactItem.ADDRESS_STATE_NONE, Address.TO,
     * Address.CC, Address.BCC
     */
    private int state;




    /**
     * the contact
     */
    private Contact contact;


    /**
     * the selected email index.
     * one of Contact.DEFAULT_EMAIL, Contact.SECONDARY_EMAIL,
     * Contact.TERTIARY_EMAIL
     *
     */
    private int selectedEmailIndex;

    /**
     * the height of this contact items
     */
    private int height;

    /**
     * the horizontal margin used while painting the contactItem
     */
    private int H_MARGIN = 5;

    /**
     * the vertical margin used while painting the contactItem
     */
    private int V_MARGIN = 3;


    /**
     * the drawer used for background painting
     */
    private Drawer drawer;

    /**
     *  the graphical theme used for colors and font
     */
    private GraphicalTheme graphicalTheme;


    /**
     * standard font
     */
    private Font font;

    /**
     * label font
     */
    private Font labelFont;

    /**
     * max label (TO:, CC:, BCC:) width in pixel, used for
     * correct spacing
     */
    private int maxLabelWidth;

    /**
     * is this contact active?
     */
    private boolean active;

    private boolean lastActive;



    /**
     * Creates a new instance of ContactItem with state=ADDRESS_STATE_NONE
     * and selectedEmail = Contact.DEFAULT_EMAIL. This is exactly the same
     * as calling ContactItem(c, ContactItem.ADDRESS_STATE_NONE, Contact.DEFAULT_EMAIL
     */
    public ContactItem(Contact c) {
        this(c, ADDRESS_STATE_NONE, Contact.DEFAULT_EMAIL);
    }

    /**
     * create a new ContactItem with given data
     *
     *
     * @param contact the contact
     * @param state one of Address.TO Address.CC Address.BCC or
     * FunCanvasContactList.ADDRESS_STATE_NONE
     * @param selectedEmail one of Contact.DEFAULT, Contact.SECONDARY_EMAIL
     * or Contact.TERTIARY_EMAIL
     */
    public ContactItem(Contact c, int state, int selectedEmail) {
        this.contact = c;
        this.state = state;
        this.selectedEmailIndex = selectedEmail;
        this.drawer = UIController.getDrawer();
        this.graphicalTheme = drawer.getGraphicalTheme();
        this.font = drawer.getGraphicalTheme().getDefaultFont();
        this.labelFont = drawer.getGraphicalTheme().getBoldFont();

    }



    /**
     * change the state of the item the the next one. <br>
     * sequence is ADDRESS_STATE_NONE -> TO -> CC -> BCC -> ADDRESS_STATE_NONE ...
     *
     * @return the state at the end of the method call, as one of
     * <ul>
     *  <li>ADDRESS_STATE_NONE</li>
     *  <li>Address.TO</li>
     *  <li>Address.CC</li>
     *  <li>Address.BCC</li>
     * </ul>
     */
    public int toNextState() {

        switch (state) {
            case ADDRESS_STATE_NONE:
                state=Address.TO;
                break;
            case Address.TO:
                state=Address.CC;
                break;
            case Address.CC:
                state=Address.BCC;
                break;
            case Address.BCC:
                state=ADDRESS_STATE_NONE;
                break;
            default:
                state=ADDRESS_STATE_NONE;
        }
        return state;
    }


    /**
     * cycle to the next email changing the value of selectedImailIndex
     * sequence is
     * DEFAULT_EMAIL -> SECONDARY_EMAIL -> TERTIARY_EMAIL -> DEFAULT_EMAIL
     */
    public void toNextEmail() {
        //we know we have at least 2 emails, and we have defaultemail
        switch (selectedEmailIndex) {

            case Contact.DEFAULT_EMAIL:
                if (! StringUtil.isNullOrEmpty( contact.getEmail_2()) ) {
                    selectedEmailIndex = Contact.SECONDARY_EMAIL;
                } else {
                    selectedEmailIndex = Contact.TERTIARY_EMAIL;
                }
                break;

            case Contact.SECONDARY_EMAIL:
                if (! StringUtil.isNullOrEmpty( contact.getEmail_3()) ) {
                    selectedEmailIndex = Contact.TERTIARY_EMAIL;
                } else {
                    selectedEmailIndex = Contact.DEFAULT_EMAIL;
                }
                break;

            default: //tertiary or unknown cases
                selectedEmailIndex = Contact.DEFAULT_EMAIL;
                break;
        }
    }

    /**
     * cycle to the previous email changing the value of selectedImailIndex.
     * Sequence is:
     * DEFAULT_EMAIL -> TERTIARY_EMAIL -> SECONDARY_EMAIL -> DEFAULT_EMAIL
     *
     */
    public void toPreviousEmail() {

        switch (selectedEmailIndex) {

            case Contact.DEFAULT_EMAIL:
                if (! StringUtil.isNullOrEmpty( contact.getEmail_3()) ) {
                    selectedEmailIndex = Contact.TERTIARY_EMAIL;
                } else {
                    selectedEmailIndex = Contact.SECONDARY_EMAIL;
                }
                break;

            case Contact.TERTIARY_EMAIL:
                if (! StringUtil.isNullOrEmpty( contact.getEmail_2()) ) {
                    selectedEmailIndex = Contact.SECONDARY_EMAIL;
                } else {
                    selectedEmailIndex = Contact.DEFAULT_EMAIL;
                }
                break;

            default: //secondary and unknown
                selectedEmailIndex = Contact.DEFAULT_EMAIL;
                break;
        }
    }

    /**
     * @return the contact in use.
     */
    public Contact getContact() {
        return contact;
    }


    /**
     * @return the address or null if no email is selected
     */
    public Address getAddress() {
        if (state==ADDRESS_STATE_NONE)
            return null;
        else {
            try {
                return contact.getAddress(state, selectedEmailIndex);
            } catch (MailException ex) {
                //TODO: better exception handling and localization
                Log.error(this, "Error getting Address from contact item");
                ex.printStackTrace();
                return null;
            }
        }
    }



    /**
     * @return the selected email as a string
     */
    private String getSelectedEmail() {
        switch (selectedEmailIndex) {
            case Contact.DEFAULT_EMAIL:
                return contact.getDefaultEmail();
            case Contact.SECONDARY_EMAIL:
                return contact.getEmail_2();
            case Contact.TERTIARY_EMAIL:
                return contact.getEmail_3();
            default:
                return contact.getDefaultEmail();
        }
    }




    // WARNING: THIS NEED TO BE REFACTORED REMOVING isActive PARAM

    /**
     * paint the ContactItem on the current graphics starting at 0,0
     * so a translate operation is often needed to correctly align the
     * graphics object
     *<br>
     * this method draws the contactItem as an item with width equals to
     * graphics.getClipWidth().
     *
     * @return the height of the painted area, this should be the same as
     * calculateheight()
     *
     * @param graphics the Graphics object to paint on.
     * @param isActive the current active / non active state of the element
     *
     */
    public int paint(Graphics graphics, boolean isActive) {


        int w = graphics.getClipWidth();//- H_MARGIN;
        int original_h = graphics.getClipHeight();
        int h = calculateHeight(isActive);
        UIController.getDrawer().drawBackground(graphics,w,h, isActive);

        if (isActive) {
            graphics.setColor(graphicalTheme.getHighlightedForegroundColor());
        } else {
            graphics.setColor(graphicalTheme.getForegroundColor());
        }

        // translate the graphics to position the icon

        //clip the rect where we'll draw the icon
        //graphics.setClip(0,0,getMaxLabelWidth()/2,font.getHeight()/3 - 2 * V_MARGIN);
        graphics.translate( w -15, 1);
        //graphics.setClip(w-15,1,12,8);
        graphics.setClip(0,0,12,8);

        drawSwitchEmailIcon(graphics);

        graphics.translate( -w +15, -1);
        graphics.setClip( 0, 0, w, original_h);


        graphics.drawLine(getMaxLabelWidth()+2*H_MARGIN,0,getMaxLabelWidth()+2*H_MARGIN,h);
        graphics.setClip(H_MARGIN,V_MARGIN, getMaxLabelWidth(),h-V_MARGIN);
        graphics.translate(H_MARGIN,V_MARGIN);
        drawLabel(graphics,w,h, isActive);
        graphics.translate(-H_MARGIN, -V_MARGIN);
        graphics.setClip(0,0,w,h);
        graphics.setClip(getMaxLabelWidth()+3*H_MARGIN,V_MARGIN,w-(getMaxLabelWidth()+3*H_MARGIN), h);
        graphics.translate(getMaxLabelWidth()+3*H_MARGIN,V_MARGIN);
        drawName(graphics,w,h, isActive);



        //back to the original graphic
        graphics.translate(-getMaxLabelWidth()-3*H_MARGIN,-V_MARGIN);
        graphics.setClip(0,0,w,original_h);

        return h;
    }



    // compute the height of the item
    private int calculateHeight(boolean isActive) {
        if (height==0 || lastActive != isActive) {
            lastActive = isActive;
            height= V_MARGIN;
            height+=Math.max(font.getHeight(), labelFont.getHeight());
            height+=V_MARGIN;
//            if ( hasMultipleEmails() ) {
            if ( isActive  ) {
                height += font.getHeight();
            }
        }
        return height;
    }

    /**
     * draws the visiblename
     */
    private void drawName(Graphics graphics, int w, int h, boolean isActive) {
        if (isActive) {
            graphics.setColor(graphicalTheme.getHighlightedForegroundColor());
        } else {
            graphics.setColor(graphicalTheme.getForegroundColor());
        }
        graphics.setFont(font);
        //TODO: use something to avoid cutstring calls every time this function
        //is called

        // if (!hasMultipleEmails() ) {
        if (!isActive) {
            drawer.drawString(UiUtils.cutString(contact.getVisibleName(),graphics.getClipWidth(),font),graphics,0,0,isActive);
        }else{
            drawer.drawString(UiUtils.cutString(contact.getVisibleName(),graphics.getClipWidth()-15,font),graphics,0,0,isActive);
            drawer.drawString(UiUtils.cutString(getSelectedEmail(),graphics.getClipWidth()-H_MARGIN,font),graphics,H_MARGIN, font.getHeight(),isActive);
        }
    }


    //  compute max label width for elements
    private int getMaxLabelWidth() {

        if (maxLabelWidth==0) {
            int w1=labelFont.stringWidth(Localization.getMessages().TO_LABEL);
            int w2=labelFont.stringWidth(Localization.getMessages().CC_LABEL);
            int w3=labelFont.stringWidth(Localization.getMessages().BCC_LABEL);
            maxLabelWidth= Math.max( w1, Math.max(w2,w3));
        }

        return maxLabelWidth;

    }

    // draw the switch email icons (arrows)
    private void drawSwitchEmailIcon(final Graphics graphics) {

        if ( hasMultipleEmails() ) {
            //graphics.setColor(0xff3333);
            graphics.fillTriangle(
                    graphics.getClipWidth() / 2 + 2 , 0    ,
                    graphics.getClipWidth(), graphics.getClipHeight() / 2,
                    graphics.getClipWidth() / 2 + 2 , graphics.getClipHeight()
                    );

            graphics.fillTriangle(
                    graphics.getClipWidth() / 2 - 2 , 0    ,
                    0, graphics.getClipHeight() / 2,
                    graphics.getClipWidth() / 2 - 2  , graphics.getClipHeight()
                    );
        }
    }

    /**
     * draws the label ("to:", "cc:","bcc:")
     */
    private void drawLabel(Graphics graphics, int w, int h, boolean isActive) {
        if (isActive)
            graphics.setColor(graphicalTheme.getHighlightedForegroundColor());
        else
            graphics.setColor(graphicalTheme.getForegroundColor());

        graphics.setFont(labelFont);
        graphics.drawString(getAddressTypeLabel(),0,0,Graphics.TOP|Graphics.LEFT);

    }


    /**
     *  @return the appropriate label given the address state
     */
    private String getAddressTypeLabel() {

        switch (state) {

            case Address.TO:
                return Localization.getMessages().TO_LABEL;

            case Address.CC:
                return Localization.getMessages().CC_LABEL;

            case Address.BCC:
                return Localization.getMessages().BCC_LABEL;

            default: //and ADDRESS_STATE_NONE
                return "";
        }

    }



    /**
     * return the current state as one of
     * Address.TO,  Address.CC, Address.BCC or ContactItem.ADDRESS_STATE_NONE
     */
    public int getState() {
        return state;
    }


    /**
     * set the state.
     *
     * @param state the state to be set. this has to be one of
     * Address.TO,  Address.CC, Address.BCC or ContactItem.ADDRESS_STATE_NONE
     * or the function will do nothing
     */
    public void setState(int state) {
        if(
                (state == Address.TO) ||
                (state == Address.CC) ||
                (state == Address.BCC) ||
                (state == ADDRESS_STATE_NONE)
                ) {
            this.state = state;
        }
    }

    /**
     * set the current contact
     * @param c the new contac
     */
    public void setContact(Contact c) {
        this.contact = c;
    }



    /**
     * check if the contact item has more than one email
     * @return true if this contact has a not null or empty
     * secondary or tertiary email
     */
    public boolean hasMultipleEmails() {
        if ( StringUtil.isNullOrEmpty( contact.getEmail_2() ) &&
                StringUtil.isNullOrEmpty( contact.getEmail_3() ) )  {
            return false;
        } else {
            return true;
        }
    }

    /**
     * reset the contact item height. since the height is cached, this method
     * should be called when the height should be recalculated
     * (e.g. if a secondary email has been added)
     */
    public void resetHeight() {
        this.height = 0;
    }

    /**
     * set the selected email.
     * value can be one of Contact.DEFAULT_EMAIL, Contac.SECONDARY_EMAIL,
     * Contac.TERTIARY_EMAIL
     */
    public void setSelectedEmailIndex(int emailIndex) {
        this.selectedEmailIndex = emailIndex;
    }

}
