/*
 * Funambol is a mobile platform developed by Funambol, Inc. 
 * Copyright (C) 2003 - 2008 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission 
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE 
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite 
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 * 
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 * 
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably 
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */
package com.funambol.mailclient;

import java.io.IOException;

import com.funambol.mailclient.config.ConfigException;
import com.funambol.mailclient.config.ConfigManager;
import com.funambol.mailclient.config.MailClientConfig;
import com.funambol.mailclient.ui.controller.AppProperties;
import com.funambol.mailclient.ui.controller.MessageHandler;
import com.funambol.mailclient.ui.controller.UIController;
import com.funambol.mailclient.sm.SyncClientConfig;
import com.funambol.push.CTPListener;
import com.funambol.push.CTPService;
import com.funambol.push.PushConfig;
import com.funambol.util.Log;
import com.funambol.util.StringUtil;

/**
 * This class is a listener for the CTP service. It uses a CTPService instance 
 * to manage all of the CTP push requests.
 */
public class CTPManager implements CTPListener {

    private static CTPManager instance = null;
    private static CTPService ctpService = null;
    private MailClientConfig mailClientConfig;

    /**
     * The contructor is private because this is a Singleton
     */
    private CTPManager() {
    }

    /**
     * Singleton Implementation: calls the contructor is the instance is null
     * otherwise return it
     * @return CTPManager instance
     */
    public static CTPManager getInstance() {
        if (instance != null) {
            return instance;
        } else {
            ctpService = CTPService.getInstance();
            return new CTPManager();
        }
    }

    /**
     * Stop the CTP service invoking the CTPService class related method
     */
    public void stopCTP() {
        ctpService.stopService();
    }

    /**
     * Start the CTP service reading the configuration from the client
     */
    public void startCTP() {
        Log.info("[CTPManager.startCTP]CTP starting");
        try {

            //
            MessageHandler messageHandler = new MessageHandler();
            // Start the CTP service (if enabled)
            //String ctp = getAppProperty("CTP");
            String ctp = UIController.appProperties.get(AppProperties.CTP);
            //#ifdef CtpPushRuntime
            //#    Log.info("enabling ctp");
            //#    ctp = "true";
            //#endif

            this.mailClientConfig = ConfigManager.getConfig();
            // setting config, ThreadPool and push notification listener
            PushConfig config = getPushConfig();
            ctpService = CTPService.getInstance();

            ctpService.setConfig(config);
            ctpService.setThreadPool(UIController.getThreadPool());
            ctpService.setCTPListener(CTPManager.getInstance());
            ctpService.setPushNotificationListener(messageHandler);

            // If CTP is disabled, then avoid starting it
            if (ctp == null || !ctp.equals("true") || this.mailClientConfig.getCtpPushStatus() == MailClientConfig.CTP_DISABLED) {
                Log.debug("CTP is disabled, not starting it");
                return;
            }
            // Start the CTP service 
            ctpService.startService();

        } catch (Exception e) {
            Log.error("No valid config found. CTPService cannot start " + e);
        }
    }

    /**
     * Restart CTP service. Stop current service Thread and create a new one.
     */
    public void restartCTP(MailClientConfig mcc) {
        Log.info("[CTPManager.restartCTP]CTP is restarting");

        this.mailClientConfig = mcc;

        PushConfig pc = null;

        try {
            pc = getPushConfig();
            if (ctpService != null) {
                ctpService.restartService(pc);
            }

        } catch (Exception ex) {
            Log.error("[CTPManager.restartCTP]Config - Error Loading PushConfig, trying a plain start");
             ex.printStackTrace();
             // config null, we start the service loading the config from rms
            this.startCTP();
       
        }


    }

    /**
     * Wrap around the CTPService.isRunning() method
     * @return true if the CTP service is running, false otherwise
     */
    public boolean isRunning() {
        return ctpService.isRunning();
    }

    /**
     * Wrap around the CTPService.setOfflineMode method
     * @param mode is the mode to be set
     */
    public void setOfflineMode(boolean mode) {
        ctpService.setOfflineMode(mode);
    }

    /**
     * get the current CTP status in string format
     * @return String formatted state
     */
    public String getCTPStringState() {
        return ctpService.getCTPStringState();
    }

    /**
     * Interface Method to notify the CTP has been Disconnected
     */
    public void CTPDisconnected() {
        Log.info("[CTPManager.CTPDisconnected]CTP service is disconnected");
    //An alarm was rescheduled in previous versions
    //As per current PRD this call must be avoided in next releases in order
    //not to create ambiguity between the CTPPush and schedule functionality
    //cm.setAlarm();
    }

    /**
     * Interface Method to notify the CTP is connecting
     */
    public void CTPConnecting() {
        Log.info("[CTPManager.CTPConnecting]CTP service is connecting...");
    }

    /**
     * Interface Method to notify the CTP is connected
     */
    public void CTPConnected() {
        Log.info("[CTPManager.CTPConnected]New CTP connection has been found");
    }

    /**
     * Interface Method to notify the CTP is autenthicating
     */
    public void CTPAuthenticating() {
        Log.info("[CTPManager.CTPAuthenticating]CTP service is authenticating...");
    }

    /**
     * Interface Method to notify the CTP is autenthicated
     */
    public void CTPAuthenticated() {
        Log.info("[CTPManager.CTPAuthenticated]CTP service is authenticated");
    }

    /**
     * Interface Method to notify the CTP is up and runnig and it is listening 
     * for SAN messages to parse. This particular implementation runs a messages 
     * sync when the CTP service is up and running so that if it was previously
     * disconnected and reconnected (i.e.: the CTP SERVER was down) the client
     * is aligned with the mailboxes last changes anyway 
     */
    public void CTPListening() {
        Log.info("[CTPManager.CTPListening]CTP service is listening...");
    //When the CTP succesfull reconnect (after the authentication process)
    //after it was down on the server
    //A sync is required in order not to lose messages that could have been
    //lost when the push service wasn't working.
    //Thus sync request has been removed after the push history was
    //implemented on server side. In order to run the sync anyway, just 
    //uncomment the following lines
    //if (!UIController.isFirstSync) {
    //    UIController.setSyncCaller(UIController.SAN_MESSAGE);
    //    UIController.sync(false);
    //}
    }

    private PushConfig getPushConfig() throws ConfigException, IOException {
        PushConfig pushConfig = new PushConfig();
        try {
            ConfigManager.load(ConfigManager.PUSH_CONFIG, pushConfig);
        } catch (Exception e) {
            // The configuration does not exist or it is invalid, we restart
            // from an empty one
            Log.info("[CTPManager.getPushConfig]CTP Config not found: " + e);
        } finally {
            // Propagate some info from the other config to the PushConfig
            Log.debug("[CTPManager.getPushConfig.finally]Retrieve Config from sync client config");
            SyncClientConfig sc = new SyncClientConfig();
            Log.debug("[CTPManager.getPushConfig.finally]Loading SyncClientConfig");
            ConfigManager.load(ConfigManager.SYNCML_CONFIG, sc);
            Log.debug("[CTPManager.getPushConfig.finally]Loading Mail account");

            Log.debug("[CTPManager.getPushConfig.finally]Mail account: " + this.mailClientConfig);

            Account account = this.mailClientConfig.getMailAccount();

            if (account == null) {
                Log.info("[CTPManager.getPushConfig.finally]" + this.mailClientConfig.toString());
                account = new Account();
            }

            Log.debug("[CTPManager.getPushConfig.finally]Reading user");
            pushConfig.setCtpUsername(account.getUser());
            Log.debug("[CTPManager.getPushConfig.finally]Reading password");
            pushConfig.setCtpPassword(account.getPassword());

            Log.debug("[CTPManager.getPushConfig.finally]Reading Server");
            String server = UIController.appProperties.get(AppProperties.CTP_SERVER);

            if (server == null) {
                Log.debug("[CTPManager.getPushConfig.finally]Extracting Server");
                server = StringUtil.extractAddressFromUrl(account.getUrl(), "http");
            } else {
                Log.debug("[CTPManager.getPushConfig.finally]CTP Port and Server");
                int pos = server.indexOf(':');
                if (pos >= 0) {
                    String portStr = server.substring(pos + 1);
                    try {
                        int port = Integer.parseInt(portStr);
                        Log.debug("Setting CTP port to: " + port);
                        pushConfig.setCtpPort(port);
                    } catch (NumberFormatException ne) {
                        Log.error("Bad ctp port parameter: " + portStr);
                    }
                    server = server.substring(0, pos);
                }

            }

            Log.debug("[CTPManager.getPushConfig.finally]CTP server is: " + server);
            pushConfig.setCtpServer(server);
            Log.debug("[CTPManager.getPushConfig.finally]Setting deviceID");
            pushConfig.setDeviceId(sc.getDeviceId());
            Log.debug("[CTPManager.getPushConfig.finally]Saving Config via ConfigManager");
            ConfigManager.save(ConfigManager.PUSH_CONFIG, pushConfig);
        }
        Log.debug("[CTPManager.getPushConfig.finally]Returning PushConfig");
        return pushConfig;
    }
}
