/*
 * Funambol is a mobile platform developed by Funambol, Inc. 
 * Copyright (C) 2003 - 2007 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission 
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE 
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite 
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 * 
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 * 
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably 
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

package com.funambol.mailclient;

import com.funambol.util.ThreadPool;
import com.funambol.util.Log;
import com.funambol.util.StringUtil;
import com.funambol.updater.Updater;
import com.funambol.updater.BasicUpdaterConfig;
import com.funambol.updater.UpdaterListener;
import com.funambol.mailclient.config.ConfigManager;
import com.funambol.mailclient.config.ConfigException;
import com.funambol.mailclient.config.MailClientConfig;
import com.funambol.mailclient.Account;

// TODO: how do we resolve these deps from framework to UI?
import com.funambol.mailclient.ui.controller.UIController;
import com.funambol.mailclient.ui.controller.AppProperties;

/**
 * This class is the client interface to the automatic update
 * library.
 */
public class UpdaterManager implements UpdaterListener {

    private static final String COMPONENT_NAME = "fjm";
    private ThreadPool threadPool;
    private Updater updater;
    private static UpdaterManager instance = null;
    private UpdaterManagerConfig config;
    private long checkInterval = (long)24 * (long)60 * (long)60 * (long)1000;   // 1 day in milliseconds
    private long reminderInterval = (long)2 * (long)60 * (long)60 * (long)1000; // 2 hours in seconds

    public static UpdaterManager getInstance() {
        if (instance == null) {
            instance = new UpdaterManager();
        }
        return instance;
    }

    class UpdaterThread extends Thread {
        public UpdaterThread() {
        }

        public void run() {
            updater.check();
        }
    }

    class UpdaterManagerConfig extends BasicUpdaterConfig {

        private long readLongProperty(String name, long defaultValue) {

            long res = defaultValue;
            String value  = UIController.appProperties.get(name);
            if (value != null && !value.equals("") ) {
                try {
                    long valueLong = Long.parseLong(value);
                } catch(NumberFormatException nfe) {
                    Log.error("Error parsing " + name + " attribute: "
                            + nfe.toString());
                }
            }
            return res;
        }


        public void load() {
            try {
                ConfigManager.load("updater.config", this);
            } catch (Exception ce) {
                Log.error("Loading updater config failed " + ce);
                // We shall create a new config
                String server = UIController.appProperties.get(AppProperties.AUTOUPDATE_URL);
                if (server == null) {
                    try {
                        MailClientConfig mailClientConfig = ConfigManager.getConfig(); 
                        Account account = mailClientConfig.getMailAccount();
                        // Guess the Updater server name from the sync server name
                        server = account.getUrl();
                        server = StringUtil.extractAddressFromUrl(server, "http");
                    } catch (Exception e) {
                        Log.error("Cannot find server information for update manager");
                    }
                }
                super.setUrl(server);
                checkInterval = readLongProperty(AppProperties.AUTOUPDATE_CHECK_INTERVAL,
                                                 checkInterval);
                setCheckInterval(checkInterval);
                reminderInterval = readLongProperty(AppProperties.AUTOUPDATE_REMINDER_INTERVAL,
                                                  reminderInterval);
                setReminderInterval(reminderInterval);
            }
        }

        public void save() {
            // save the configuration to RMS
            try {
                ConfigManager.save("updater.config", this);
            } catch (Exception ce) {
                Log.error("Saving updater config failed " + ce);
            }
        }
    }

    private UpdaterManager() {
        String version = ConfigManager.getClientVersion();
        this.threadPool = UIController.getThreadPool();
        config = new UpdaterManagerConfig();
        config.load();
        updater = new Updater(config, version, COMPONENT_NAME);
        updater.setListener(this);
    }

    public void check() {
        // The update check is executed in a separate thread
        UpdaterThread updThread = new UpdaterThread();
        threadPool.startThread(updThread);
    }

    public void mandatoryUpdateAvailable(String version) {
        // In this client an update is available only if we have a new
        // version and a valid url.
        String url = getDownloadUrl();
        if (url != null && url.length() > 0) {
            UIController.notifyNewMandatoryUpdate(version);
        }
    }

    public void optionalUpdateAvailable(String version) {
        // In this client an update is available only if we have a new
        // version and a valid url.
        String url = getDownloadUrl();
        if (url != null && url.length() > 0) {
            UIController.notifyNewOptionalUpdate(version);
        }
    }

    public void updateLastReminder() {
        long now = System.currentTimeMillis();
        updater.setLastReminder(now);
    }

    public void userSelectedLater() {
    }

    public void userSelectedSkip() {
        updater.setSkip();
    }

    public boolean isUpdateAvailable() {
        // In this client an update is available only if we have a new
        // version and a valid url.
        String url = getDownloadUrl();
        if (url != null && url.length() > 0) {
            return updater.isUpdateAvailable();
        } else {
            return false;
        }
    }

    public String getDownloadUrl() {
        return config.getDownloadUrl();
    }

    public void setUrl(String url) {
        config.setUrl(url);
        // Since a new url has been set, we also reset our timers to force a new
        // check and possibly a new reminder
        config.setLastReminder(0);
        config.setLastCheck(0);
        config.save();
    }
}

