/*
 * Funambol is a mobile platform developed by Funambol, Inc. 
 * Copyright (C) 2003 - 2007 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission 
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE 
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite 
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 * 
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 * 
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably 
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */
package com.funambol.mailclient.config;

import com.funambol.mailclient.Account;
import com.funambol.mailclient.MailFilter;
import com.funambol.mailclient.sm.SyncClientConfig;
import com.funambol.mailclient.Funambol;
import com.funambol.mailclient.ui.controller.AppProperties;
import com.funambol.mailclient.ui.controller.UIController;
import com.funambol.updater.BasicUpdaterConfig;
import com.funambol.push.PushConfig;

import com.funambol.syncml.spds.SourceConfig;
import com.funambol.storage.AbstractRecordStore;
import com.funambol.storage.ObjectStore;
import com.funambol.storage.Serializable;

import com.funambol.util.DateUtil;
import com.funambol.util.Log;
import com.funambol.util.StringUtil;

import java.io.IOException;
import java.io.DataInputStream;
import java.util.Date;
import java.util.Hashtable;

import javax.microedition.rms.RecordStore;
import javax.microedition.rms.RecordStoreNotOpenException;
import javax.microedition.rms.RecordStoreException;
import javax.microedition.midlet.MIDlet;

/**
 * The configuration manager is resposible for the store and retrieve
 * of the client config data in the Storage.
 *
 * Usage:
 *
 * mailConfig = ConfigManager.getConfig(); // to obtain the mail config params
 *
 * // To load and save a source config
 * ConfigManager.load("source.name", SourceConfig);
 * ConfigManager.save("source.name", SourceConfig);
 */
public class ConfigManager {
    
    
    public static final int MAIL_CLIENT_CONFIG    = 1;
    public static final int MAIL_SOURCE_CONFIG    = 2;
    public static final int CONTACT_SOURCE_CONFIG = 3;
    public static final int SYNCML_CONFIG         = 4;
    public static final int PUSH_CONFIG           = 5;
    public static final int UPDATER_CONFIG        = 6;
    
    private static final String CONFIGNAME = "MailConfig";
    private static final int HI_MSG_NUMBER = 350;
    private static final int LO_MSG_NUMBER = 50;
    
    // Config storage
    private static ObjectStore confstore = new ObjectStore();
    // Mail client config
    private static MailClientConfig mailClientConfig;
    
    private static String clientName;
    private static String clientVendor;
    private static String clientVersion;
    
    private static int FILTER_WINDOW = 10;
    
    private static int storageSize;
    
    private static int maxMessageNumber = LO_MSG_NUMBER;
    
    private static boolean firstRun = false;
    
    public static String appInfoAttr;
    
    /** Temporary property (not serialized) for switching
     * from compressed sync to uncompressed sync
     */
    public static boolean _tmpCompress;
    
    //------------------------------------------------------------- Constructors
    
    /**
     * This class is used by calling static members, prevent
     * callers from creating instances.
     */
    private ConfigManager() {
    }
    
    //----------------------------------------------------------- Public methods
    
    /**
     * Load the Serializable object in the config store, with the given name
     */
    public static Serializable load(String name, Serializable obj)
    throws ConfigException, IOException {
        
        int id = getId(name);
        return load(id, obj);
    }
    
    public static void save(String name, Serializable obj)
    throws ConfigException, IOException {
        
        int id = getId(name);
        save(id, obj);
    }
    
    
    public static Serializable load(int configId, Serializable obj)
    throws ConfigException, IOException {
        try {
            confstore.open(CONFIGNAME);
            obj = confstore.retrieve(configId, obj);
            if (obj == null) {
                throw new ConfigException("Non existing config");
            }
        } catch (RecordStoreException e) {
            Log.info("Can't load " + configId + ": " + e.toString());
            throw new ConfigException(e.toString());
        }
        return obj;
    }
    
    /**
     * Save the Serializable object in the config store, with the given name.
     *
     * @return true if the config has been saved for the first time.
     */
    public static void save(int configId, Serializable obj)
    throws ConfigException, IOException {
        
        try {
            confstore.create(CONFIGNAME);
            confstore.store(configId, obj);
        } catch (RecordStoreException e) {
            Log.error("Can't save " + configId + ": " + e.toString());
            throw new ConfigException(e.toString());
        }
    }
    
    /**
     * Return the cached mail config.
     *
     * @return the cached MailClientConfig.
     *
     * @throws ConfigException if the config is not present
     */
    public static MailClientConfig getConfig()
    throws ConfigException{
        
        if(mailClientConfig == null) {
            // init configuration before using this method
            throw new ConfigException("Error: Configuration not yet initialized");
        }
        
        return mailClientConfig;
    }
    
    /**
     * Return the cached or stored mail config. If not present
     * loads it from JAD or create a default one.
     *
     * @return the MailClientConfig stored or cached or from JAD.
     *
     * @throws ConfigException if the config is not present
     * @throws IOException if an error occurred during the deserialization
     */
    public static MailClientConfig getConfig(Funambol midlet)
    throws ConfigException, IOException {
        MailClientConfig tmp = new MailClientConfig();
        try {
            if(mailClientConfig == null) {
                mailClientConfig = (MailClientConfig) load(MAIL_CLIENT_CONFIG, tmp);
                Log.info("MailClientConfig from Storage\n");
            }
        } catch (ConfigException e) {
            // First Run. Loading configuration from JAD or create a default one
            Log.info("First run. Configuration from JAD");
            midlet.setSplashPhase("Preparing for first run...", 0);
            firstRun = true;
            recoverConfig(true);
        } catch (IOException e) {
            // Error retreiving configuration from storage
            // Loading configuration from JAD or create a default one
            Log.error("Error loading configuration from storage");
            midlet.setSplashPhase("Initialization error, recovering...", 0);
            recoverConfig(false);
        } finally {
            // Settings always from Jad, never stored in Storage
            clientName = UIController.appProperties.get(
                    AppProperties.MIDLET_NAME);
            Log.info("ClientName: " + clientName);
            clientVersion = UIController.appProperties.get(AppProperties.MIDLET_VERSION);
            Log.info("ClientVersion: " + clientVersion);
            clientVendor = UIController.appProperties.get(AppProperties.MIDLET_VENDOR);
            Log.info("ClientVendor: " + clientVendor);
            //#ifdef isBlackberry
            //# boolean flagCompress = true;
            //#else
            boolean flagCompress = StringUtil.getBooleanValue(
                    UIController.appProperties.get(AppProperties.ENABLE_COMPRESS));
            //#endif
            Log.info("Compress from Jad: " + "[" + flagCompress + "]");
            mailClientConfig.enableCompress(flagCompress);
            _tmpCompress = flagCompress;
            
            // Disable date filter only if present and set to false
            String dateFilter = UIController.appProperties.get(AppProperties.ENABLE_DATE_FILTER);
            if(dateFilter != null) {
                if (StringUtil.equalsIgnoreCase("false", dateFilter)) {
                    Log.info("DateFilter from Jad: [false]");
                    mailClientConfig.enableDateFilter(false);
                }
            }
            
            
            appInfoAttr = UIController.appProperties.get(AppProperties.APP_INFO_ATTR);
            /**if (appInfoAttr!=null && !appInfoAttr.startsWith("Moto")) {
             * maxMessageNumber=HI_MSG_NUMBER;
             * }**/
            
            //#ifndef LowMsgNum
            maxMessageNumber=HI_MSG_NUMBER;
            //#endif
        }
        
        return mailClientConfig;
    }
    
    
    
    /**
     * Save the cached mail config, if previously loaded.
     */
    public static void saveConfig() throws ConfigException, IOException {
        if(mailClientConfig != null) {
            save(MAIL_CLIENT_CONFIG, mailClientConfig);
        }
    }
    
    /**
     * Specifies if this is the first run (no config was pre-saved)
     */
    public static boolean getFirstRun() {
        return firstRun;
    }
    
    /**
     * Save the new mail config, and cache it.
     */
    public static void saveConfig(MailClientConfig mcc)
    throws ConfigException, IOException {
        if(mcc != null) {
            save(MAIL_CLIENT_CONFIG, mcc);
            mailClientConfig = mcc;
        }
    }
    
    /** Get the name of the Mail Client from jad*/
    public static String getClientName() {
        return clientName;
    }
    
    /** Get the version of the mail client from jad */
    public static String getClientVersion() {
        return clientVersion;
    }
    
    /** Get the name of the Vendor from jad*/
    public static String getClientVendor() {
        return clientVendor;
    }
    
    /**
     * Accessor method to retrieve Storage size
     * @return int value of rms storage size.
     */
    public int getStorageSize() {
        return storageSize;
    }
    
    public static int getMaxMessageNumber() {
        return maxMessageNumber;
    }
    
    /** This method converts the configuration version 102 to 103.
     * It is for backward compatibily only and it can be removed in future
     * releases.
     * The method loads the configuration and rewrites it in the new format. The
     * main purpose it to replace the NamedObjectStore with an ObjectStore for
     * configuration storage.
     *
     * @return true if upgrade was necessary.
     */
    public static boolean upgrade() {
        
        try {
            confstore.open(CONFIGNAME);
        } catch (Exception e) {
            // if the config does not exist, this is not un upgrade
            return false;
        }
        
        try {
            int size = confstore.size();
            
            DataInputStream in = confstore.retrieveBytes(1);
            int firstInt = in.readInt();
            
            // This is kind of tricky, but as we have no global configuration
            // version information we are forced to play dirty
            if (firstInt != 4) {
                // This is not an hashtable, upgrade is reguired
                return false;
            }
            
            Log.info("Configuration needs to be upgraded");
            
            MailClientConfig mcc = new MailClientConfig();
            SourceConfig msc = new SourceConfig();
            SourceConfig csc = new SourceConfig();
            SyncClientConfig scc = new SyncClientConfig();
            
            for(int i=0; i<4; i++ ){
                String name = in.readUTF();
                int recordId = in.readInt();
                if (name.equals("MailConfig")) {
                    confstore.retrieve(recordId, mcc);
                } else if (name.equals("source.mail")) {
                    confstore.retrieve(recordId, msc);
                } else if (name.equals("source.contact")) {
                    confstore.retrieve(recordId, csc);
                } else if (name.equals("syncml.config")) {
                    confstore.retrieve(recordId, scc);
                } else {
                    Log.error("Cannot upgrade configuration. Unknown config " + name);
                    return false;
                }
            }
            
            initConfigurationStorage(mcc, msc, csc, scc, null, null);
            return true;
            
        } catch (Exception ex) {
            // This should never happen... but if it does we must delete the
            // configuration and ask the user to create a new one
            Log.error("Cannot upgrade configuration");
            initConfigurationStorage(null, null, null, null, null, null);
            return false;
        }
    }
    
    
    //--------------------------------------------------------- Private Methods
    
    // retrieve account and configuration settings from JAD attributes
    private static MailClientConfig getConfigFromJad() {
        Log.info("MailClientConfig from JAD or default\n");
        
        MailClientConfig configFromJad = new MailClientConfig();
        
        configFromJad.setMailAccount(getMailAccountFromJad());
        
        int logLevel = getLogLevelFromJAD();
        configFromJad.setLogLevel(logLevel);
        
        String pollInterval = UIController.appProperties.get(AppProperties.POLL_INTERVAL_ATTR);
        if (pollInterval != null && !pollInterval.equals("") ) {
            try {
                int pollIntervalInt = Integer.parseInt(pollInterval);
                configFromJad.setPollInterval(pollIntervalInt);
                
            } catch(NumberFormatException nfe) {
                Log.error("Error parsing pollInterval attribute: "
                        + nfe.toString());
            }
        }
        
        boolean flag = false;
        String enableScheduler = UIController.appProperties.get(AppProperties.ENABLE_SCHEDULER_ATTR);
        if (enableScheduler != null && !enableScheduler.equals("") ) {
                flag = StringUtil.getBooleanValue(enableScheduler);
                configFromJad.enableScheduler(flag);
            
        }
        
        flag = StringUtil.getBooleanValue(
                UIController.appProperties.get(AppProperties.ENABLE_SMS_LISTENER_ATTR));
        //#ifdef isBlackberry
        //# configFromJad.enableSmsListener(true);
        //#else
        configFromJad.enableSmsListener(flag);
        //#endif
        flag = StringUtil.getBooleanValue(
                UIController.appProperties.get(AppProperties.ENABLE_DELETE_PROPAGATION));
        configFromJad.enableDeletePropagation(flag);
        
        return configFromJad;
    }
    
    private static int getLogLevelFromJAD() {
        String jadStr = UIController.appProperties.get(AppProperties.LOGLEVEL_ATTR);
        //#ifdef isBlackberry
        //# jadStr = "DEBUG";
        //#endif
        if ((jadStr == null) || (jadStr.equals(""))) {
            return Log.DISABLED;
            
        } else if (StringUtil.equalsIgnoreCase(jadStr, "DISABLED")) {
            return Log.DISABLED;
            
        } else if (StringUtil.equalsIgnoreCase(jadStr, "ERROR")) {
            return Log.ERROR;
            
        } else if (StringUtil.equalsIgnoreCase(jadStr, "INFO")) {
            return Log.INFO;
            
        } else if (StringUtil.equalsIgnoreCase(jadStr, "DEBUG")) {
            return Log.DEBUG;
            
        } else if (StringUtil.equalsIgnoreCase(jadStr, "TRACE")) {
            return Log.TRACE;
            
        } else {
            return Log.ERROR;
            
        }
    }
    
    // get account settings from JAD
    private static Account getMailAccountFromJad() {
        Account mailAccount = new Account();
        
        String user = UIController.appProperties.get(AppProperties.USER_ATTR);
        if (StringUtil.isNullOrEmpty(user)) {
            Log.info("No user defined in JAD file");
        } else {
            Log.info("User got from JAD: " + "[" + user + "]");
            mailAccount.setUser(user);
        }
        
        String url = UIController.appProperties.get(AppProperties.URL_ATTR);
        if (StringUtil.isNullOrEmpty(url)) {
            Log.info("No url defined in JAD file");
        } else {
            Log.info("Url got from JAD: " + "[" + url + "]");
            mailAccount.setUrl(url);
        }
        
        String password = UIController.appProperties.get(AppProperties.PASSWORD_ATTR);
        if (StringUtil.isNullOrEmpty(password)) {
            Log.info("No password defined in JAD file");
        } else {
            Log.info("Password got from JAD: " + "[" + password + "]");
            mailAccount.setPassword(password);
            
        }
        
        String name = UIController.appProperties.get(AppProperties.NAME_ATTR);
        if (StringUtil.isNullOrEmpty(name)) {
            Log.info("No name defined in JAD file");
        } else {
            Log.info("Name got from JAD: " + "[" + name + "]");
            mailAccount.setName(name);
        }
        
        String address = UIController.appProperties.get(AppProperties.ADDRESS_ATTR);
        if (StringUtil.isNullOrEmpty(address)) {
            Log.info("No address defined in JAD file");
        } else {
            Log.info("Address got from JAD: " + "[" + address + "]");
            mailAccount.setAddress(address);
        }
        
        String remoteURI = UIController.appProperties.get(AppProperties.REMOTEURI_ATTR);
        if (StringUtil.isNullOrEmpty(remoteURI)) {
            Log.info("No remoteURI defined in JAD file");
        } else {
            Log.info("RemoteURI got from JAD: " + "[" + remoteURI + "]");
            mailAccount.setRemoteURI(remoteURI);
        }
        
        String pimRemoteURI = UIController.appProperties.get(AppProperties.PIM_REMOTEURI_ATTR);
        if (StringUtil.isNullOrEmpty(pimRemoteURI)) {
            Log.info("No pimRemoteURI defined in JAD file");
        } else {
            Log.info("pimRemoteURI got from JAD: " + "[" + pimRemoteURI + "]");
            mailAccount.setPimRemoteURI(pimRemoteURI);
        }
        
        return mailAccount;
    }
    
    /**
     * Get the device ID specifird in the JAD, if present.
     * The device ID can be obtained from the JAD to avoid the
     * generation of a new dev-id each time the application is installed
     * without keeping the old configurration.
     * It is mainly intented for debug purposes, but the portal can also
     * set this value to have always the same id for a registered device.
     */
    private static void initDeviceIdFromJad()
    throws ConfigException, IOException {
        //Get SyncClientConfig for SyncML specific parameters...
        SyncClientConfig syncClientConf = new SyncClientConfig();
        
        //get the device ID from the JAD (if available)
        String devID = UIController.appProperties.get(AppProperties.DEVID_ATTR);
        if (StringUtil.isNullOrEmpty(devID)) {
            Log.info("No device ID defined in JAD file");
        } else {
            Log.info("Device ID got from JAD: " + "[" + devID + "]");
            // ... set device ID in the SyncClientConfig object...
            syncClientConf.setDeviceId(devID);
            save(SyncClientConfig.CONFIGNAME, syncClientConf);
        }
    }
    
    private static int getId(String name) {
        
        int configId = 0;
        if (name.equals("MailConfig")) {
            configId = MAIL_CLIENT_CONFIG;
        } else if (name.equals("source.mail")) {
            configId = MAIL_SOURCE_CONFIG;
        } else if (name.equals("source.contact")) {
            configId = CONTACT_SOURCE_CONFIG;
        } else if (name.equals("syncml.config")) {
            configId = SYNCML_CONFIG;
        } else if (name.equals("push.config")) {
            configId = PUSH_CONFIG;
        } else if (name.equals("updater.config")) {
            configId = UPDATER_CONFIG;
        }
        return configId;
    }
    
    private static int loadRmsSize() {
        int size = 512*1024;
        try {
            AbstractRecordStore rs = AbstractRecordStore.openRecordStore(MailClientConfig.NAME, true);
            size = rs.getSizeAvailable();
            rs.closeRecordStore();
        } catch (RecordStoreNotOpenException ex) {
            Log.error("RecordStoreNotOpenException: " + ex.getMessage());
            ex.printStackTrace();
        } catch (RecordStoreException ex) {
            Log.error("RecordStoreException: " + ex.getMessage());
            ex.printStackTrace();
        }
        return size;
    }
    
    private static void recoverConfig(boolean firstrun)
    throws ConfigException, IOException {
        
        // If we are here then we need to re-init the config because either it
        // does not exist or it is unreadable
        initConfigurationStorage(null,null,null,null,null,null);
        
        mailClientConfig = getConfigFromJad();
        int size =0;
        if (firstrun) {
            size = loadRmsSize();
        } else {
            size = 512*1024;
        }
        mailClientConfig.setRmsStoreSize(size);
        
        saveConfig(mailClientConfig);
        // init the SyncClientConfig with the device id from jad, if present
        // otherwise, a default one will be generated.
        initDeviceIdFromJad();
    }
    
    /**
     * This method creates the necessary record for storing the J2ME client
     * configuration.
     *
     * @param mcc the MailClientConfig or null if it does not exist (the record
     * will be created empty and the load method will report is as non existing)
     *
     * @param msc the mail SourceConfig or null if it does not exist (the record
     * will be created empty and the load method will report is as non existing)
     *
     * @param csc the contact SourceConfig or null if it does not exist (the record
     * will be created empty and the load method will report is as non existing)
     *
     * @param scc the SyncClientConfig or null if it does not exist (the record
     * will be created empty and the load method will report is as non existing)
     */
    private static void initConfigurationStorage(MailClientConfig mcc,
            SourceConfig msc,
            SourceConfig csc,
            SyncClientConfig scc,
            PushConfig pcfg,
            BasicUpdaterConfig ucfg) {
        try {
            confstore.close();
            AbstractRecordStore.deleteRecordStore(CONFIGNAME);
        } catch (Exception e) {
            // If the config store does not exist, this is OK
        }
        
        try {
            // Now recreate it empty
            confstore.create(CONFIGNAME);
            // And write four records (one for each config)
            if (mcc == null) {
                confstore.createEmptyRecord();
            } else {
                confstore.store(mcc);
            }
            if (msc == null) {
                confstore.createEmptyRecord();
            } else {
                confstore.store(msc);
            }
            if (csc == null) {
                confstore.createEmptyRecord();
            } else {
                confstore.store(csc);
            }
            if (scc == null) {
                confstore.createEmptyRecord();
            } else {
                confstore.store(scc);
            }
            if (pcfg == null) {
                confstore.createEmptyRecord();
            } else {
                confstore.store(pcfg);
            }
            if (ucfg == null) {
                confstore.createEmptyRecord();
            } else {
                confstore.store(ucfg);
            }
        } catch (Exception e) {
            Log.error("Cannot init configuration");
        }
    }
}

