/*
 * Funambol is a mobile platform developed by Funambol, Inc. 
 * Copyright (C) 2003 - 2007 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission 
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE 
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite 
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 * 
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 * 
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably 
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

package com.funambol.mailclient.config;

import com.funambol.util.MailDateFormatter;
import java.io.DataOutputStream;
import java.io.DataInputStream;
import java.io.IOException;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;

import java.util.Date;

import com.funambol.mailclient.Account;
import com.funambol.mailclient.MailFilter;
import com.funambol.mail.Store;
import com.funambol.mail.StoreFactory;

import com.funambol.storage.Serializable;
import com.funambol.util.DateUtil;
import com.funambol.util.Log;



/**
 * Configuration data for the Mail Client: account, polling time etc.
 *
 */
public class MailClientConfig implements Serializable {
    
    //---------------------------------------------------------------- Constants
    /** This is the default polling interval*/
    
    /** The name of the config record */
    public static final String NAME = "MailConfig" ;
    
    /** ctp status */
    public static final int CTP_ENABLED = 0;
    public static final int CTP_DISABLED = 1;
    
    /* This field contains the previous versions of the configuration data
     * from which the backward compatibility is guaranteed
     */
    private static final int VERSION_101 = 101;
    private static final int VERSION_102 = 102;
    private static final int VERSION_103 = 103;
    private static final int VERSION_104 = 104;
    // v. 105 adds ctp status settings
    private static final int VERSION_105 = 105;

    //Adds the deleted open message settings
    private static final int VERSION_106 = 106;

    /* This field contains the version of the configuration data */
    private static final int VERSION = VERSION_106 ;
    
    //--------------------------------------------------------------- Attributes
    
    /** The configuration version loaded from the store */
    private long version ;
    
    /** Mail params */
    private Account mailAccount;
    
    /** Mail Filter */
    private MailFilter mailFilter;
    
    /** Enable scheduled sync */
    private boolean enableScheduler = false;
    /** Polling interval, in minutes */
    private int pollInterval = 0;
    /** Enable Sms listener */
    private boolean enableSmsListener;
    /** Enable Delete command propagation*/
    private boolean enableDeletePropagation;
    /** Log level*/
    private int logLevel;
    /** Time of daily polling */
    private long nextTimeAlarm;
    /** Client Server Communications Compress */
    private boolean compress;
    
    //--------------------------------- Fields added in version 101
    
    /** Date of First installation */
    private long startFilter;
    
    // Number of days to remove for the date of starting for filter
    private int DAYS_TO_REMOVE = 3;
    
    private int rmsStoreSize;
    
    //--------------------------------- Fields added in version 102
    
    private String timeZone;
    
    private static boolean timeZoneUpdated; //  NOT Serialized
    
    //---------------------------------- Fields added in version 103
    
    private boolean soundNotification;
    private boolean vibrateNotification;
    private int     storeVersion;
    private boolean syncOnStartup;
    
    //---------------------------------- Fields added in version 104
    private boolean deleteConfirmation;

    //---------------------------------- Fields added in version 105
    /** ctp push status*/
    private int ctpPushStatus = CTP_ENABLED;
  
    //---------------------------------- Fields added in version 106
    private int deleteOpenMessage = 1;
    
    //------------------------------------------------------------- Constructors
    
    /**
     * Default constructor. Sets default configuration values.
     */
    public MailClientConfig() {
        version = VERSION;
        
        // Mail parameters
        mailAccount = new Account();
        Date date = new DateUtil().toMidnight().removeDays(DAYS_TO_REMOVE).getTime();
        
        // Default filter parameters
        
        
        mailFilter = new MailFilter(2*1024,   // size:2K
            date,   // from date
            false,      // headers only
            false);     // without attachments
        
        
        // Push parameters
        enableSmsListener = false;
        enableDeletePropagation = false;
        logLevel = Log.INFO ;
        compress = true;
        // Version 101
        startFilter = date.getTime();
        rmsStoreSize = 0;               // XXX: use another default?
        timeZoneUpdated = false;
        soundNotification = true;
        vibrateNotification = true;
        
        // By default we have the latest version of the store version,
        // because this is the case when we're not upgrading
        storeVersion = Store.LATEST_VERSION;
        StoreFactory.getStore().setVersion(storeVersion);
        syncOnStartup = true;

        deleteConfirmation = true;
        
        deleteOpenMessage = 1;
    }
    
    //----------------------------------------------------------- Public methods
    
    public long getStartFilter() {
        return startFilter;
    }
    
    public void setStartFilter(long startFilter) {
        this.startFilter = startFilter;
    }
    
    public Account getMailAccount() {
        return this.mailAccount;
    }
    
    public void setMailAccount(Account account) {
        this.mailAccount = account;
    }
    
    /** Return the currently configured mail filter.  */
    public MailFilter getMailFilter() {
        return mailFilter;
    }
    
    /** Set a new mail filter.  */
    public void setMailFilter(MailFilter mailFilter) {
        this.mailFilter = mailFilter;
    }
    
    /** Returns the current status of the scheduler. */
    public boolean isSchedulerEnabled() {
        return enableScheduler;
    }
    
    /** Enables or disables the scheduler */
    public void enableScheduler(boolean enableScheduler) {
        this.enableScheduler = enableScheduler;
    }
    
    /** This value is not used at the moment by the client. */
    public boolean isSmsListenerEnabled() {
        return enableSmsListener;
    }
    
    /** This value is not used at the moment by the client. */
    public void enableSmsListener(boolean enableSmsListener) {
        this.enableSmsListener = enableSmsListener;
    }
    
    /** Sets the log level. See com.funambol.util.Log for valid levels. */
    public void setLogLevel(int logLevel) {
        this.logLevel = logLevel;
    }
    
    /** Gets the log level. See com.funambol.util.Log for valid levels. */
    public int getLogLevel() {
        return logLevel;
    }
    
    /** Sets the polling interval, in seconds. */
    public void setPollInterval(int pollInterval) {
        this.pollInterval = pollInterval;
    }
    
    /** Returns the polling interval, in seconds. */
    public int getPollInterval() {
        return this.pollInterval;
    }
    
    /** Returns true if the communication with the server is compressed. */
    public boolean isCompress() {
        return compress;
    }
    
    /** Enables or disables compression in the communication with the server */
    public void enableCompress(boolean enable) {
        compress = enable;
    }
    
    /** Enables or disables the date filter */
    public void enableDateFilter(boolean enable) {
        if(enable) {
            mailFilter.setDate(new Date(startFilter));
        } else {
            mailFilter.setDate(null);
        }
    }
    
    /** Returns the current status of the delete propagation on server */
    public boolean isEnableDeletePropagation() {
        return enableDeletePropagation;
    }
    
    /** Enables or disables the delete propagation on server */
    public void enableDeletePropagation(boolean deletePropagation) {
        this.enableDeletePropagation = deletePropagation;
    }
    
    /** Gets the time of the next wakeup for the daily scheduler */
    public long getNextTimeAlarm() {
        return nextTimeAlarm;
    }
    
    /** Sets the time of the next wakeup for the daily scheduler */
    public void setNextTimeAlarm(long nextTimeAlarm) {
        this.nextTimeAlarm = nextTimeAlarm;
    }
    
    /** Gets the RMS store size for the midlet. */
    public int getRmsStoreSize() {
        return rmsStoreSize;
    }
    
    /** Sets the RMS store size for the midlet. */
    public void setRmsStoreSize(int rmsStoreSize) {
        this.rmsStoreSize = rmsStoreSize;
    }
    
    public void setTimeZone(String timeZoneToSet){
        this.timeZone = timeZoneToSet;
        MailDateFormatter.setTimeZone(timeZone);
        this.timeZoneUpdated = true;
    }
    
    public String getTimeZone(){
        return timeZone;
    }
    
    public boolean  getTimeZoneUpdated(){
        return timeZoneUpdated;
    }
    
    //-------------------------------- getters and setters for sound and vibrate
    //-------------------------------- notifications added in version 103
    
    /**
     * @return true if sound notification is enabled
     */
    public boolean isSoundNotificationEnabled() {
        return soundNotification;
    }
    /**
     * @return true if vibrate notification is enabled
     */
    public boolean isVibrateNotificationEnabled() {
        return vibrateNotification;
    }
    /**
     * enable or disable sound notifications on a new email
     */
    public void setSoundNotification(boolean enabled) {
        this.soundNotification = enabled;
    }
    
    /**
     * enable or disable vibrate notifications on a new email
     */
    public void setVibrateNotification(boolean enabled) {
        this.vibrateNotification= enabled;
    }
    
    /**
     * @return store version (format). @see Store for more details.
     */
    public int getStoreVersion() {
        return storeVersion;
    }
    
    /**
     * @return store version (format). @see Store for more details.
     */
    public void setStoreVersion(int version) {
        storeVersion = version;
    }

    public boolean isSyncOnStartupEnabled() {
        return syncOnStartup;
    }
    
    public void setSyncOnStartup(boolean enabled) {
        syncOnStartup = enabled;
    }

    // getters and setters for sound and vibrate
    // added in version 104
    public boolean getDeleteConfirmation() {
        return deleteConfirmation;
    }

    public void setDeleteConfirmation(boolean deleteConfirmation) {
        this.deleteConfirmation = deleteConfirmation;
    }
    
    
    //getters and setters for push status
    // added in version 105
    public int getCtpPushStatus() {
        return ctpPushStatus;
    }

    public void setCtpPushStatus(int ctpPushStatus) {
        this.ctpPushStatus = ctpPushStatus;
    }

    public void setDeleteOpenMessage(int deleteOpenMessage) {
        this.deleteOpenMessage = deleteOpenMessage;
    }
    
    public int getDeleteOpenMessage() {
        return this.deleteOpenMessage;
    }
    
    
    /**
     * @return current configuartion as string
     */
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("Configuration version: ").append(this.version);
        sb.append("\nMailAccount: \n").append(this.mailAccount);
        sb.append("\nMailFilter: ").append(this.mailFilter);
        sb.append("\nEnableScheduler: ").append(this.enableScheduler);
        sb.append("\nEnableSmsListener: ").append(this.enableSmsListener);
        sb.append("\nLogLevel: ").append(this.logLevel);
        sb.append("\nPollInterval: ").append(this.pollInterval);
        sb.append("\nNextTimeAlarm: "
                +
                ((nextTimeAlarm!=0L) ?
                    MailDateFormatter.dateToUTC(new Date(nextTimeAlarm)): "null")
                    );
        sb.append("\nEnableDeletePropagation: ")
        .append(this.enableDeletePropagation);
        sb.append("\nCompress: ").append(this.compress);
        // Ver.101
        sb.append("\nStartFilter: ").append(MailDateFormatter.dateToUTC(new Date(startFilter)));
        sb.append("\nRmsSize: ").append(rmsStoreSize);
        sb.append("\nSoundNotification: ").append(soundNotification);
        sb.append("\nVibrateNotification: ").append(vibrateNotification);
        sb.append("\nSyncOnStartup: ").append(syncOnStartup);
        sb.append("\nStore version: ").append(storeVersion);
        sb.append("\nCTPStatus: ").append(ctpPushStatus == 0 ? "enabled" : "disabled");
        sb.append("\ndeleteOpenMessage: ").append(deleteOpenMessage);
        return sb.toString();
    }

    public MailClientConfig clone() {
        MailClientConfig copy = new MailClientConfig();
        ByteArrayOutputStream ostream = new ByteArrayOutputStream();
        DataOutputStream os = new DataOutputStream(ostream);
        try {
            this.serialize(os);
            ByteArrayInputStream istream = new ByteArrayInputStream(ostream.toByteArray());
            DataInputStream is = new DataInputStream(istream);
            copy.deserialize(is);
            return copy;
        } catch (Exception e) {
            return null;
        }
    }

    //---------------------------------------------- Serializable implementation
    
    /**
     * Write object fields to the output stream.
     * @param out Output stream
     * @throws IOException
     */
    public void serialize(DataOutputStream out) throws IOException {
        
        
        out.writeInt(VERSION);
        
        out.writeUTF(mailAccount.getUrl());
        out.writeUTF(mailAccount.getUser());
        out.writeUTF(mailAccount.getPassword());
        out.writeUTF(mailAccount.getName());
        out.writeUTF(mailAccount.getAddress());
        out.writeUTF(mailAccount.getRemoteURI());
        out.writeUTF(mailAccount.getPimRemoteURI());
        
        out.writeBoolean(enableScheduler);
        out.writeInt(pollInterval);
        out.writeBoolean(enableSmsListener);
        
        out.writeBoolean(enableDeletePropagation);
        
        out.writeInt(logLevel);
        out.writeLong(nextTimeAlarm);
        
        out.writeInt(mailFilter.getSize());
        
        // Handle the case of null date
        Date d = mailFilter.getDate();
        out.writeLong((d != null) ? d.getTime() : 0L);
        
        out.writeBoolean(mailFilter.downloadAttachments());
        out.writeBoolean(mailFilter.isHeadersOnly());
        // Ver.101
        out.writeLong(startFilter);
        out.writeInt(rmsStoreSize);
        
        //Ver.103
        out.writeBoolean(soundNotification);
        out.writeBoolean(vibrateNotification);
        out.writeInt(storeVersion);
        out.writeBoolean(syncOnStartup);

        //Ver.104
        out.writeBoolean(deleteConfirmation);
        
        //Ver.105
        out.writeInt(getCtpPushStatus());

        //Ver.106
        out.writeInt(deleteOpenMessage);
        
    }
    
    /**
     * Read object field from the input stream.
     * @param in Input stream
     * @throws IOException
     */
    public void deserialize(DataInputStream in) throws IOException {
        boolean enableFirstInstall = true;
        int savedVer = in.readInt();
        
        // If the config stored belongs to a newer version than the
        // current one (the user went back to an older version, use
        // the defaults.
        if (savedVer > VERSION) {
            Log.error("This app is older then the config stored.");
            Log.error("Can't handle forward compatibility, using defaults.");
            return;
        }
        
        version = savedVer;
        
        mailAccount.setUrl(in.readUTF());
        mailAccount.setUser(in.readUTF());
        mailAccount.setPassword(in.readUTF());
        mailAccount.setName(in.readUTF());
        mailAccount.setAddress(in.readUTF());
        mailAccount.setRemoteURI(in.readUTF());
        mailAccount.setPimRemoteURI(in.readUTF());
        
        enableScheduler = in.readBoolean();
        pollInterval = in.readInt();
        enableSmsListener = in.readBoolean();
        
        enableDeletePropagation = in.readBoolean();
        
        logLevel = in.readInt();
        nextTimeAlarm = in.readLong();
        
        mailFilter.setSize(in.readInt());
        // Handle the case of null date
        long d = in.readLong();
        mailFilter.setDate((d != 0L) ? new Date(d) : null);
        
        mailFilter.enableAttachmentsDownload(in.readBoolean());
        mailFilter.enableHeadersOnly(in.readBoolean());
        
        // Handle backward compatibilty
        
        if (savedVer >= VERSION_101) {
            startFilter = in.readLong();
            rmsStoreSize = in.readInt();
        }
        
        // Before version 103 we had no store version,
        // but the store format (version) was 101
        if (savedVer <= VERSION_102) {
            storeVersion = Store.VERSION_101;
            StoreFactory.getStore().setVersion(storeVersion);
        }
        
        // if version is at least 103 we can read
        // the values, otherwise we use the default
        // values (true)
        if (savedVer >= VERSION_103) {
            soundNotification = in.readBoolean();
            vibrateNotification = in.readBoolean();
            storeVersion = in.readInt();
            StoreFactory.getStore().setVersion(storeVersion);
            syncOnStartup = in.readBoolean();
        }

        // Properties introduced in version 104
        if (savedVer >= VERSION_104) {
            deleteConfirmation = in.readBoolean();
        }
        
        // Properties introduced in version 105
        if (savedVer >= VERSION_105) {
            ctpPushStatus=in.readInt();
        }
        
        if (savedVer >= VERSION_106) {
            deleteOpenMessage=in.readInt();
        }
        
    }

}

