/*
 * Funambol is a mobile platform developed by Funambol, Inc. 
 * Copyright (C) 2003 - 2007 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission 
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE 
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite 
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 * 
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 * 
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably 
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

package com.funambol.mailclient.mm;

import com.funambol.mail.BodyPart;
import com.funambol.mail.MessageFlags;
import com.funambol.mail.Multipart;
import com.funambol.mailclient.syncml.FlagQueue;
import java.io.UnsupportedEncodingException;

import java.util.Hashtable;
import java.util.Date;
import java.util.Vector;


import com.funambol.mailclient.config.ConfigManager;
import com.funambol.mailclient.config.ConfigException;
import com.funambol.mail.Message;
import com.funambol.mail.Folder;
import com.funambol.mail.Store;
import com.funambol.mail.StoreFactory;
import com.funambol.mail.MailException;

import com.funambol.mail.MIMEProcessor;
import com.funambol.mail.MIMEFormatter;


import com.funambol.mailclient.config.MailClientConfig;
import com.funambol.util.Log;
import com.funambol.util.XmlException;
import com.funambol.util.XmlUtil;
import com.funambol.util.ChunkedString;
import com.funambol.util.MailDateFormatter;
import com.funambol.util.QuotedPrintable;
import com.funambol.util.StringUtil;

/**
 * This class is used to manage local messages.The MessageManager is responsible
 * for adding/updating/deleting messages. All these operations can be performed
 * on the client and then propagated to the server, or originated on the server
 * and propagated to the client.
 * Message (@see Message)is the basic entity to represent an email. Message is
 * handled almost completely by the mail API, but there is one information in
 * Message which is left to the client to manage. The key is a concept for the
 * client and Messages created by the API leave it empty.
 * The MessageManager is responsible for creating all the Messages in the
 * client. Message should never be created without using the MessageManager. In
 * this context the MessageManager acts as a Message factory for the client.
 * Any time the MessageManager creates a Message it sets the key to a proper
 * value, guaranteeing that no Message is without key. The key is used to
 * save/load messages, therefore it is crucial that each Message has this
 * information properly set.
 * From a message key it is always possible to derive both the message folder
 * and the message record id in the underlying store.
 * Other parts of the client should not make any assumption on how the LUID is
 * formed.
 * The MessageManager has three different set of methods to handle messages:
 * 
 * 1) generic methods to operate on the client messages
 * 2) methods invoked to create/update/delete messages upon a user action. In
 *    these cases the UI notifies that something changed and this information
 *    must be sent to the server on the next synchronization
 * 3) methods invoked to create/update/delete messages upon a server request. In
 *    these cases the information in the system (e.g. RMS) is updated. The
 *    MessageManager does not call back any method, because notifications to the
 *    UI are sent via SyncListener by the SyncManager.
 *
 */
public class MessageManager {
    
    //---------------------------------------------------------------- Constant
    public static final int INBOX  = 0 ;
    public static final int OUTBOX = 1 ;
    public static final int DRAFTS = 2 ;
    public static final int SENT   = 3 ;
    //public static final int TRASH   = 4 ; trash is hidden
    
    //-------------------------------------------------------------- Attributes
    private Store msgStore = null;
    
    private static MessageManager mm = null;

    private Folder[] rootFolders;
    
    //------------------------------------------------------------ Constructors
    
    /**
     * Default constructor.
     */
    private MessageManager() {
        try {
            msgStore = StoreFactory.getStore();
        } catch (Exception e) {
            Log.error(this, "Cannot get store: " + e.toString());
            msgStore = null;
        }
    }

    
    public static MessageManager getInstance() {
        if (mm == null) {
            mm = new MessageManager();
        }
        return mm;
    }
    
    //---------------------------------------------------------- Public Methods
    
    
    
    /**
     * Re-initialize the mail the store.
     *
     * @param reset if true, erase and re-create all the main folders.
     *
     * @throws MailException in case of error (re)creating folders.
     */
    public void initStore(boolean reset) throws MailException {
        msgStore.init(reset);
        // clear queue of flags if re-initialization is needed
        if (reset) {
            FlagQueue.clear();
        }
    }
    
    /**
     * Get the list of the top level folders from the store.
     */
    public Folder[] getRootFolders() {
        if (rootFolders == null) {
            computeRootFolders();
        }
        return rootFolders;
    }

    

    ////////////////////////////////////////////////////////////////////////////
    // Methods invoked by the client to apply changes to messages. The changes
    // are originated on the client side. All these methods update the
    // information and keep tracks of what happened to notify the server on the
    // next synchronization
    ////////////////////////////////////////////////////////////////////////////
   
    /**
     * Delete a message from a folder
     * @param msg the message to be deleted
     */
    public void deleteMessage(Message msg) throws ConfigException, MailException {
        removeFromFlagQueue(msg);
        if ((ConfigManager.getConfig().isEnableDeletePropagation())) {
            updateMessageFlag(msg, MessageFlags.DELETED, true);
        }
        msg.getParent().deleteMessage(msg);

    }

    /**
     * Append a message to a folder. The message key is set after this
     * operation. The method throws an exception if the underlying folder
     * throws an exception.
     *
     * @param msg the message
     * @param folder the folder
     *
     * @throws MailException if the underlying folder throws an exception
     */
    public void appendMessageToFolder(Message msg, Folder folder) throws MailException {
        folder.appendMessage(msg);
        msg.setKey(makeLuid(msg));
    }

    /**
     * Append a message to a folder. The message key is set after this
     * operation. The method throws an exception if the underlying folder
     * throws an exception.
     *
     * @param msg the message
     * @param folder the folder id
     *
     * @throws MailException if the underlying folder throws an exception
     */
    public void appendMessageToFolder(Message msg, int folderIdx) throws MailException {
        Folder folder = getFolder(folderIdx);
        appendMessageToFolder(msg, folder);
    }
   
    
    /**
     * remove a folder and recreate it empty
     * @param  folder the folder to be removed (as one of <ul>
     * <li>MessageManager.INBOX</li>
     * <li>MessageManager.OUTOX</li>
     * <li>MessageManager.DRAFTS</li>
     * <li>MessageManager.SENT</li></ul>
     * @throws MailException if we have problem recreating the folder
     */
    public void emptyFolder(int folderId) throws MailException {
        
        
        String folderName = "";
        switch (folderId) {
            case INBOX:
                folderName = Store.INBOX;
                break;
            case OUTBOX:
                folderName = Store.OUTBOX;
                break;
            case DRAFTS:
                folderName = Store.DRAFTS;
                break;
            case SENT:
                folderName = Store.SENT;
                break;
        }
        if (!"".equals(folderName)) {
            msgStore.removeFolder(folderName);
            msgStore.createFolder(folderName);
        }
    }

    /**
     * Updates a message which is already stored in a folder (the folder
     * information must be valid)
     *
     * @param msg the message
     * @throws MailException if the underlying folder throws an exception
     */
    public void updateMessage(Message msg) throws MailException {
        Folder folder = msg.getParent();
        folder.updateMessage(msg);
    }
    
    /**
     * Sets Message Flags after a messages one or more of its property flag
     * @param msg the message to be flagged
     * @param flag is the flag to be added to the message <code>msg</code>
     */
    public void updateMessageFlag(Message msg, int flag, boolean value)
    throws MailException {
        
        // The flag is already set, nothing to change.
        boolean current = msg.getFlags().isSet(flag);
        if(current == value) {
            Log.debug("Flag: "+flag+" already set");
            return;
        }
        
        //Creates MessageFlags to put into updated items HashTable
        MessageFlags flags = new MessageFlags(msg.getFlags());
        flags.setFlag(flag, value);
        
        // OPENED or FLAGGED flags must be set into the current
        // msg and saved immediately, the others will be set after the sync.
        switch (flag) {
            case MessageFlags.DELETED:
                updateMessageFlags(msg, flags);
                break;
            case MessageFlags.ANSWERED:
            case MessageFlags.DRAFT:
            case MessageFlags.FORWARDED:
            case MessageFlags.PARTIAL:
                //Call update for SyncML
                updateMessageFlags(msg, flags);
                break;
            case MessageFlags.OPENED:
            case MessageFlags.FLAGGED:
                //Call update for SyncML
                updateMessageFlags(msg, flags);
                //Save the updated flags message into the folder
                msg.getFlags().setFlag(flag, value);
                Folder msgFolder = msg.getParent();
                msgFolder.appendMessage(msg);
                break;
            default:
                break;
                
        }
    }
       
    
    /**
     * Update message flags Lists to be read by MailSyncSource
     * with this implementation the flags are managed only for the Inbox folder
     *
     *@param msg the the message to be put in the queue
     *@param mf the flags to be set
     */
    public void updateMessageFlags(Message msg, MessageFlags mf) {
        if(msg.getParent().getName().equals(Store.INBOX)) {
            FlagQueue.put(msg.getKey(), mf);
        }
        /* Leave commented here: could be useful to move here the FlagQueue merging logic
        if(msg.getParent().getName().equals(Store.INBOX)) {
            Log.debug("[Flagqueue.put()] New Flags");
            Log.debug("[Flagqueue.put()]: " + mf.toString());

            MessageFlags currentFlags = FlagQueue.get(msg.getKey());
            if (currentFlags!=null) {
                Log.debug("[MessageManager] Old Flags");
                Log.debug("[MessageManager]: " + currentFlags.toString());
                currentFlags.merge(mf);
            } else {
               currentFlags=mf;
            }

                        
            Log.debug("[MessageManager] Updating FlagQueue...");
            FlagQueue.put(msg.getKey(), currentFlags);
            Log.debug("[MessageManager] FlagQueue Updated");
            Log.debug("[MessageManager] new message has: " +
                    currentFlags.toString()
                    );
            
        }*/
    }
    
    /**
     * Remove message flags from the queue
     * with this implementation the flags are managed only for the Inbox folder
     *
     *@param msg the the message to be put in the queue
     *@param mf the flags to be set
     */
    public void removeFromFlagQueue(Message msg) {
        if(msg.getParent().getName().equals(Store.INBOX)) {
            FlagQueue.remove(msg.getKey());
        }
    }
    
    
    
    
    ////////////////////////////////////////////////////////////////////////////
    // Methods invoked by the client to apply changes to messages. The changes
    // are originated on the server side.
    ////////////////////////////////////////////////////////////////////////////
 
    /**
     * Creates a new Message from a SyncML byte stream. The message is also
     * saved in its folder.
     *
     * @param parentName parent folder name
     * @param guid       server side message id
     * @param emailData  byte stream of the SyncML email
     *
     * @return a new Message object with the proper key defined or
     *         null if either the folder is invalid or the byte stream cannot be
     *         parsed properly.
     *
     * @throws XmlException  if the byte stream cannot be parsed
     * @throws MailException if the message cannot be saved in the folder
     */
    public Message createNewMessage(String parentName, String guid,
                                    byte[] emailData)
    throws XmlException, MailException {

        // Dump memory stats
        Log.stats("Begin of MessageManager.createNewMessage");

        String folderName = folderFromParent(parentName);
        if(folderName == null) {
            Log.error("[handleEmailItem]Invalid parent from server: "
                      +parentName);
            // TODO FIXME throw an exception?
            return null;
        }

        String luid;
        Message msg = parseMessage(emailData);
        
        if (msg != null) {
            String msgid = msg.getMessageId();
            Folder f = msgStore.getFolder(folderName);
            Log.info("[MM.addMessage]Adding message " + msgid +
                     " to folder " + f.getName());
            f.appendMessage(msg);
            // We set the luid to be more information than just the recordId
            // Our LUID is "folder"/"recordid"
            msg.setKey(makeLuid(msg));
        }

        // Dump memory stats
        Log.stats("End of MessageManager.createNewMessage");

        return msg;
    }

    /**
     * Creates a new Message from a SyncML byte stream. The message is NOT 
     * saved in any folder. (@see createNewMessage(String, String, byte[])).
     * The message must already be present in the client because the method
     * expects an existing valid luid. The newly created message will have such
     * a luid.
     * Note that after this method is invoked we have two messages with the same
     * LUID but different contents. It is up to the client to preserve the
     * system integrity.
     * In the J2ME email client this method is used in case of "get more" to
     * retireve a full message which is not saved in the RMS but just fetched to
     * be displayed.
     *
     * @param luid       client side message id
     * @param emailData  byte stream of the SyncML email
     *
     * @return a new Message object with the proper key defined or
     *         null if the byte stream cannot be parsed properly.
     *
     * @throws XmlException  if the byte stream cannot be parsed
     */
    public Message updateMessageContent(String luid, byte[] emailData)
    throws XmlException {

        Message msg = parseMessage(emailData);
        if (msg != null) {
            // We set the luid to be more information than just the recordId
            // Our LUID is "folder"/"recordid"
            msg.setKey(luid);
        }
        return msg;
    }

    /**
     * Updates an existing message by redefining the sent date and/or the
     * received date. The message is also written into the folder that gets
     * updated immediately.
     *
     * @param sentDate sent date. If null the message sent date is left
     *        unchanged
     *
     * @param receivedDate received date. If null the message received date is
     *        left unchanged
     *
     * @throws MailException if the folder update fails
     */
    public void updateMessageDates(Message msg, Date sentDate, Date receivedDate)
    throws MailException {
        String luid = msg.getKey();
        String parent = folderFromKey(luid);
        Folder f = msgStore.getFolder(parent);
        String recordId = msgIdFromKey(luid);

        if (sentDate != null) {
            msg.setSentDate(sentDate);
        }
        if (receivedDate != null) {
            msg.setReceivedDate(receivedDate);
        }
                            
        f.updateMessage(msg);
    }

    /**
     * Updates an existing message by redefining its flags. The message is also
     * written into the folder that gets updated immediately. The method accepts
     * the SyncML complete description of a message, but the content is ignored
     * and only the flags are decoded to update the flags.
     *
     * @param parentName parent folder name
     * @param luid message client id
     * @param emailData SyncML byte stream describing the message
     *
     * @throws MailException if the folder update fails
     * @throws XmlException if the byte stream is not SyncML complaint
     */
    public MessageFlags updateMessageFlags(String parentName, String luid,
                                           byte[] emailData)
    throws XmlException, MailException {

        String folderName = folderFromParent(parentName);
        if(folderName == null) {
            Log.error("[handleEmailItem]Invalid parent from server: "
                      +parentName);
            // TODO FIXME throw an exception?
            return null;
        }

        MessageFlags flags = parseFlags(emailData);

        // Update the message in the store
        try {
            Folder f = msgStore.getFolder(parentName);

            // Flag update only.
            Log.debug("[updateMessageFlags]updating flags for: " + luid);

            String recordId = msgIdFromKey(luid);
            Message saved = f.getMessage(recordId);
            // We reset the luid to be more information from what the RMSStore
            // can define. Our LUID is "folder"/"recordid"
            if (saved!=null) {
                //if the message has been deleted this would generate NPE
                saved.setKey(luid);
                saved.setFlags(flags);
                f.updateMessage(saved);
            }
            return flags;

        } catch (MailException me) {

            Log.error("[updateMessageFlags]Error saving message: " + luid 
                    + " " + me.toString() );
            me.printStackTrace();
            throw me;
        }
    }

    /**
     * Delete an existing message which is actually removed from the folder it
     * belongs to.
     * When we try to delete a message in the Outbox and such an operation
     * fails, we check because the failure may be due to the fact that the
     * server is notifying a succesfull sent. On a succesfull sent the server
     * moves messages from Outbox to Sent. On the client we move from Outbox to
     * Sent when the server accepts the message to be sent (assuming that sooner
     * or later the message will be sent). Therefore we must ignore the
     * subsequent move operation.
     *
     * @param luid message local id
     * @throws MailException if the folder update fails
     */
    public void deleteMessage(String luid) throws MailException {
            
        String parent = folderFromKey(luid);
        Folder f = msgStore.getFolder(parent);
        String recordId = msgIdFromKey(luid);
        try {
            f.deleteMessage(recordId);
        } catch (MailException exc) {
            // If the message to delete was in the outbox
            // it just means it has been sent. We move these
            // messages from outbox to sent as soon as the server inform us that
            // it accepted the item to send, therefore we do not have this email
            // in the outbox any longer. We can ignore the exception in such a
            // case.
            // TODO: would it be even safer to check that the message is now in
            // the sent folder?
            if (!(exc.getCode() == MailException.MESSAGE_DELETE_ERROR &&
                  f.getName().equals(Store.OUTBOX))) {
                throw exc;
            }
        }
    }

    /**
     * This method should be called on a message that has been accepted by the
     * server as to be sent. The method moves the message from local Outbox to
     * local Sent and updates the folders.
     *
     * @param luid local message id to be moved from outbox to sent
     *
     * @throws MailException if one of the operations on the folder fails
     *         (delete from Outbox or add to the Sent)
     */
    public void messageEffectivelySent(String luid) throws MailException {
        // Move the message to the sent folder
        Folder sent   = msgStore.getFolder(Store.SENT);
        Folder outbox = msgStore.getFolder(Store.OUTBOX);

        Message msg = getMessage(outbox, luid);
        msg.getFlags().setFlag(MessageFlags.TX_SENDING, false);
        msg.getFlags().setFlag(MessageFlags.TX_SENT, true);
        msg.reloadMessage();
        outbox.deleteMessage(msg);
        // Update the luid and the message parent folder
        msg.setParent(sent);
        msg.setKey(makeLuid(msg));
        // And finally append the message into the sent folder
        sent.appendMessage(msg);
    }

    /**
     * Get the basic informations for the messages contained in
     * a given folder.
     * This method is not just a wrapper for the Folder.getMsgHeaders as it sets
     * the key of all the Messages returned. In the client this method should
     * always be used instead of the Folder counterpart.
     *
     * @param folderName is the folder name
     * @return an array of Message held by this folder
     * @throws MailException if errors occur retrieving the folder content
     */
    public Message[] getMessagesInFolder(String folderName) throws MailException {

        // Dump memory stats
        //Log.stats("Beginning of MessageManager.getMessagesInFolder");

        Folder folder = msgStore.getFolder(folderName);
        /*Message[] msgs = folder.getMsgHeaders();
        int count = msgs.length;
        
        for(int i=0; i<count; i++) {
            Message msg = msgs[i];
            msg.setKey(makeLuid(msg));
        }

        // Dump memory stats
        Log.stats("End of MessageManager.getMessagesInFolder");
        
        return msgs;*/
        
        return getMessagesInFolder(folder);
    }



    /**
     * Get the basic informations for the messages contained in
     * a given folder.
     * This method is not just a wrapper for the Folder.getMsgHeaders as it sets
     * the key of all the Messages returned. In the client this method should
     * always be used instead of the Folder counterpart.
     *
     * @param folder is the folder
     * @return an array of Message held by this folder
     * @throws MailException if errors occur retrieving the folder content
     */
    public Message[] getMessagesInFolder(Folder folder) throws MailException {

        // Dump memory stats
        Log.stats("Beginning of MessageManager.getMessagesInFolder");

        
        Message[] msgs = folder.getMsgHeaders();
        int count = msgs.length;
        
        for(int i=0; i<count; i++) {
            Message msg = msgs[i];
            msg.setKey(makeLuid(msg));
        }

        // Dump memory stats
        Log.stats("End of MessageManager.getMessagesInFolder");
        
        return msgs;
    }

    
    
    
    
    
    /**
     * Return a single message fetched from its folder.
     *
     * @param key the message key
     * @return the Message
     */
    public Message getMessage(String key) {
        String folderName = folderFromKey(key);
        Folder folder     = msgStore.getFolder(folderName);
        return getMessage(folder, key);
    }

    /**
     * Return a single message fetched from its folder.
     *
     * @param folder the folder containing the message
     * @param key the message key
     * @return the Message
 
     */
    public Message getMessage(Folder folder, String key) {
        String recordId = msgIdFromKey(key);
        // Get the message
        Message msg = folder.getMessage(recordId);
        // Set Message key
        if (msg!=null) {
            //if the message has been previously deleted this generates NPE
            msg.setKey(makeLuid(msg));
        }
        return msg;
    }
    
    /**
     * Returns the first message in the given folder.
     *
     * @return the first message if it exists, null otherwise
     * @throws MailException if an existing message cannot be read, or if the
     * folder is not in the store
     */
    public Message getFirstMessage(Folder folder) throws MailException {
        Message msg = folder.getFirstMessage();
        // Set Message key
        if (msg != null) {
            msg.setKey(makeLuid(msg));
        }
        return msg;
    }

    /**
     * Returns the next message in the given folder.
     *
     * @return the next message if it exists, null otherwise
     * @throws MailException if an existing message cannot be read, or if the
     * folder is not in the store
     */
    public Message getNextMessage(Folder folder) throws MailException {
        Message msg = folder.getNextMessage();
        // Set Message key
        if (msg != null) {
            msg.setKey(makeLuid(msg));
        }
        return msg;
    }

    /**
     * Returns the Folder corresponding to the given folder id.
     * If the folder id does not correspond to any real folder, the method may
     * throw an OutOfBoundArrayException.
     *
     * (@see MessageManager.INBOX, @see MessageManager.OUTBOX,
     *  @see MessageManager.DRAFT, @see MessageManager.SENT)
     *
     *  @param folder the folder id
     */
    public Folder getFolder(int folder) {
        Folder tmpRootFolders[] = getRootFolders();
        return tmpRootFolders[folder];
    }
    

    /**
     * Get the message Id from the key.
     *
     * @param key is the message key
     * @return is the record id
     */
    public String msgIdFromKey(String key) {
        int pos = key.indexOf('/');
        
        if(pos != -1) {
            return key.substring(pos+1);
        }
        return key;
    }

    /**
     * Get the folder name from the key.
     *
     * @param key is the message key
     * @return the folder name
     */
    public String folderFromKey(String key) {
        if(key == null) {
            return null;
        }
        int pos = key.indexOf('/');
        
        if(pos != -1) {
            return key.substring(0, pos);
        }
        return null;
    }

    /**
     * Dumps an email flags into SyncML format.
     *
     * @param msg is the email message
     * @return the SyncML format
     */
    public byte[] formatFlags(MessageFlags flags) {
        StringBuffer out = new StringBuffer();
        out.append("<Email>\n");
        formatFlags(flags, out);
        out.append("</Email>\n");
        return out.toString().getBytes();
    }

    /**
     * Dumps an email into SyncML format.
     *
     * @param msg is the email message
     * @return the SyncML format
     */
    public byte[] format(Message msg) {
        MessageFlags flags = msg.getFlags();
        StringBuffer out = new StringBuffer();
        out.append("<Email>\n");

        formatFlags(flags, out);

        String utcdate = null;
        Date received = msg.getReceivedDate();
        Date created  = msg.getSentDate();
        Date modified = msg.getSentDate();
        
        if(received != null){
            utcdate = MailDateFormatter.dateToUTC(received);
            XmlUtil.addElement(out, TREC_TAG, utcdate);
        }
        if(created != null){
            utcdate = MailDateFormatter.dateToUTC(created);
            XmlUtil.addElement(out, TCRE_TAG, utcdate);
        }
        if(modified != null) {
            utcdate = MailDateFormatter.dateToUTC(modified);
            XmlUtil.addElement(out, TMOD_TAG, utcdate);
        }
        
        MIMEFormatter mimef = new MIMEFormatter();
        out.append("<emailitem>\n<![CDATA[");
        // Append MIME representation of emailItem to out.
        mimef.format(msg, out);
        out.append("]]&gt;\n</emailitem>\n");
        out.append("</Email>\n");

        return out.toString().getBytes();
    }

    
    /**
     * Dumps flags and msgid of an email into SyncML format.
     *
     * @param msg is the email message
     * @return the SyncML format with only flags and msgid
     */
    public byte[] formatMinimalInfo(Message msg) {
        MessageFlags flags = msg.getFlags();
        StringBuffer out = new StringBuffer();
        out.append("<Email>\n");

        // append the the flags
        formatFlags(flags, out);
        
        MIMEFormatter mimef = new MIMEFormatter();
        out.append("<emailitem>\n<![CDATA[");
        
        // Append MIME representation of message id 
        mimef.formatOnlyMsgID(msg, out);
        out.append("]]&gt;\n</emailitem>\n");
        out.append("</Email>\n");

        return out.toString().getBytes();
    }

    
    
    
    /**
     * Creates a new Folder. At the moment we do not really create any local
     * folder. We have a set of predefined folders (Inbox, Outbox, Draft and
     * Sent). Any folder propagated by the server is just ignored.
     *
     * @param syncmlStream syncml byte stream
     * @return the Folder
     * @throws XmlException if the byte stream cannot be parsed
     */
    public Folder createFolder(byte syncmlStream[]) throws XmlException {
        return parseFolder(syncmlStream);
    }

 
    //--------------------------------------------------------- Private Fields
    // These fields are SyncML tags for email description

    private static final String EMAIL_TAG = "Email";
    private static final String READ_TAG = "read";
    private static final String FORW_TAG = "forwarded" ;
    private static final String REPL_TAG = "replied" ;
    private static final String TREC_TAG = "received" ;
    private static final String TCRE_TAG = "created" ;
    private static final String TMOD_TAG = "modified" ;
    private static final String DELE_TAG = "deleted" ;
    private static final String FLAG_TAG = "flagged" ;
    private static final String ITEM_TAG = "emailitem" ;
    private static final String EXT_TAG  = "Ext" ;

    // These fields are SyncML tags for folder description
    private static final String FOLDER_TAG = "Folder";
    private static final String NAME_TAG = "name";
    private static final String CREATED_TAG = "created";
    private static final String ROLE_TAG = "role";

   

    //--------------------------------------------------------- Private Methods

    /**
     * Utility function to parse an email message flags (the email is a SyncML
     * byte stream)
     *
     * @param syncmlData is the SyncML byte stream
     * @return the MessageFlags
     */
    private MessageFlags parseFlags(byte[] syncmlData)
    throws XmlException, MailException {
        ChunkedString data = new ChunkedString(new String(syncmlData));
        ChunkedString xml = XmlUtil.getTagValue(data, EMAIL_TAG);
        return parseFlags(xml);
    }

     /**
     * Utility function to parse an email message flags (the email is a SyncML
     * chunked string)
     *
     * @param xml is the SyncML byte stream tokenized
     * @return the MessageFlags
     */
    private MessageFlags parseFlags(ChunkedString xml)
    throws XmlException, MailException {

        MessageFlags flags = new MessageFlags();

        flags.setFlag(MessageFlags.OPENED,    getXmlFlag(xml, READ_TAG));
        flags.setFlag(MessageFlags.FORWARDED, getXmlFlag(xml, FORW_TAG));
        flags.setFlag(MessageFlags.ANSWERED,  getXmlFlag(xml, REPL_TAG));
        flags.setFlag(MessageFlags.DELETED,   getXmlFlag(xml, DELE_TAG));
        flags.setFlag(MessageFlags.FLAGGED,   getXmlFlag(xml, FLAG_TAG));

        return flags;
    }

    /**
     * This method parses a SyncML byte stream describing an email and builds
     * the corresponding Message object.
     *
     * @param syncmlData email in SyncML byte stream format
     * @return the Message
     * @throws XmlException if the byte stream cannot be parsed
     */
    private Message parseMessage(byte[] syncmlData)
    throws XmlException {

        // Dump memory stats
        Log.stats("Beginning of MessageManager.parseMessage");

        ChunkedString data = null;
        try {
            // The encoding must be set to UTF-8 
            // in order to read the symbols like "euro" 
            data = new ChunkedString(new String(syncmlData, "UTF-8"));
        } catch (UnsupportedEncodingException uee) {
            uee.printStackTrace();
            Log.error("[MessageManager] Cannot convert email data");
        }
        
        ChunkedString xml = XmlUtil.getTagValue(data, EMAIL_TAG);
        data = null;

        MessageFlags flags = parseFlags(xml);
       
        Date received = getXmlDate(xml, TREC_TAG);
        Date created =  getXmlDate(xml, TCRE_TAG);
        Date modified = getXmlDate(xml, TMOD_TAG);

        int itemPos = XmlUtil.getTag(xml, ITEM_TAG);
        if(itemPos > -1) {
            int begin = xml.indexOf(">", itemPos);
            if(xml.charAt(begin-1) == '/') {
                Log.debug("EmailData.parse: empty emailitem.");
                return null;
            }
            int end = xml.indexOf("</"+ITEM_TAG+">", begin);
            if(begin == -1 || end == -1) {
                Log.error("[EmailData] error parsing EmailItem");
                throw new XmlException("Can't parse email item");
            }
            // FIXME: implement the CDATA handling in the Xml parser or use
            //        kxml
            int pos = xml.indexOf("<![CDATA[", begin);
            if (pos != -1) {
                begin = pos + 8;
                end = xml.indexOf("]]>", begin);
                
                if (end == -1) {
                    Log.error("[EmailData] can't find CDATA end.");
                    throw new XmlException("Can't find CDATA end.");
                }
            }
            ChunkedString item = xml.substring(begin+1, end);
            
            Hashtable attr = XmlUtil.getTagAttributes(xml, ITEM_TAG);
            String encoding = (String)attr.get("enc");
            // Try with Uppercase for 3.0 connector bug
            if(encoding == null) {
                encoding = (String)attr.get("ENC");
            }
            if(encoding != null) {
                Log.debug("EmailData.parse: encoding from server= " + encoding);
                if(StringUtil.equalsIgnoreCase(encoding,"quoted-printable")) {
                    // TODO: optimize this. Since the connector now does
                    //       not send QP to the client, we will not fall
                    //       into this case.
                    String qp = item.toString();
                    qp = QuotedPrintable.decode(qp.getBytes(), "UTF-8");
                    item = new ChunkedString(qp);
                }
            }
            // 
            MIMEProcessor mp = new MIMEProcessor();
            Message msg = mp.parseMailMessage(item);
            
            msg.setFlags(flags);

            // Process the Funambol extensions for filtering. Example:
            // <Ext><XNam>x-funambol-body</XNam><XVal>15000</XVal></Ext> 
            // <Ext><XNam>x-funambol-attach-n</XNam><XVal>1</XVal></Ext> 
            // <Ext><XNam>x-funambol-attach</XNam><XVal>att1.txt</XVal><XVal>10000</XVal></Ext>
            // XXX: we are now considering x-funambol-body only
            Vector extList = XmlUtil.getTagValues(xml, EXT_TAG);
            
            Multipart multipart = null;
            
            
            
            for(int i=0, l=extList.size(); i<l; i++) {
                ChunkedString ext = (ChunkedString)extList.elementAt(i);
                try {
                    ChunkedString xnam = XmlUtil.getTagValue(ext, "XNam");
                    if(xnam.equals("x-funambol-body")) {
                        String xval = XmlUtil.getTagValue(ext, "XVal").toString();
                        if (xval == null) {
                            Log.info("Incomplete header");
                        }
                        msg.getFlags().setFlag(MessageFlags.PARTIAL, true);
                        //break;
                    }

                  
//ifndef isBlackberry
                    
                    if(xnam.equals("x-funambol-attach-n")){
                        if(msg.isMultipart()){
                            multipart = (Multipart)msg.getContent();
                        }else{
                            multipart = new Multipart();
                            BodyPart bodyPartText = new BodyPart();
                            bodyPartText.addHeader(BodyPart.CONTENT_TRANSFER_ENCODING, "8bit");
                            bodyPartText.setContent(msg.getTextContent());
                            multipart.addBodyPart(bodyPartText);
                        }
                        
                        
                    }
        
                    if(xnam.equals("x-funambol-attach")){
                        Vector vec = XmlUtil.getTagValues(ext, "XVal");
                        BodyPart bodyPartAttach = new BodyPart();
                        bodyPartAttach.setDisposition(BodyPart.CD_ATTACHMENT);
                        bodyPartAttach.addHeader(BodyPart.CONTENT_TRANSFER_ENCODING, "8bit");
                        String name = ((ChunkedString)vec.elementAt(0)).toString();
                        if (name.equals("null")){
                            name = "Att.html";
                        }
                        bodyPartAttach.setFileName(name);
                        bodyPartAttach.setSize(((ChunkedString)vec.elementAt(1)).toString());
                        if (vec.size() == 4){
                            bodyPartAttach.setContentType(((ChunkedString)vec.elementAt(2)).toString());
                            
                            String url = checkUrl(((ChunkedString)vec.elementAt(3)).toString());
                            bodyPartAttach.setAttachUrl(url);
                        }
                        multipart.addBodyPart(bodyPartAttach); 
                    }
                            
//endif   
                } catch (XmlException xe) {
                    xe.printStackTrace();
                    Log.error("[EmailData.parse]Invalid Ext field. Ignoring it");
                }
            }
            
            if(multipart != null){
                msg.setContent(multipart);
                Log.debug("Message is multipart : "+msg.isMultipart());
            }
            // Dump memory stats
            Log.stats("End of MessageManager.parseMessage");

            return msg;
        }
        return null;
    }
    
     private String checkUrl(String url) {
        if(url.startsWith("/")){
            try{
                String syncUrl = ConfigManager.getConfig().getMailAccount().getUrl();
                int index = syncUrl.indexOf('/',7);
                String prefixUrl = syncUrl.substring(0, index);
                return prefixUrl+url;                
            } catch (ConfigException cex) {
                cex.printStackTrace();
                Log.error(this,"Extracting attachments url from ext "+cex.toString());
                return null;
            }
            
        }else{
            return url;
        }
         
         
    }

    /**
     * Parse a SyncML byte stream representing a folder.
     *
     * @param syncmlData syncml byte stream
     * @return the Folder
     * @throws XmlException if the SyncML item is not formed properly
     */
    private Folder parseFolder(byte[] syncmlData) throws XmlException {
        ChunkedString xml = new ChunkedString(new String(syncmlData));

        String name;
        String role = XmlUtil.getTagValue(xml, ROLE_TAG).toString();
        //TODO: be careful while doing turkish localization: there are 
        //two different lowercase 'i'  in turkey!
        
        if (role.toLowerCase().equals(Store.INBOX.toLowerCase())) {
            Log.debug("Folder Inbox from server");
            name = Store.INBOX;
        }
        else if (role.toLowerCase().equals(Store.OUTBOX.toLowerCase())) {
            Log.debug("Folder Outbox from server");
            name = Store.OUTBOX;
        }
        else if (role.toLowerCase().equals(Store.DRAFTS.toLowerCase())) {
            Log.debug("Folder Drafts from server");
            name = Store.DRAFTS;
        }
        else if (role.toLowerCase().equals(Store.SENT.toLowerCase())) {
            Log.debug("Folder Sent from server");
            name = Store.SENT;
        }
        else {
            throw new XmlException("Invalid folder from server:"+role);
        }

         //Date created = MailDateFormatter.parseUTCDate(XmlUtil.getTagValue(xml, CREATED_TAG).toString());

         Folder folder = new Folder(name, msgStore);
         return folder;
    }


    /**
     * Utility method to check if a tag is found in a chunked string
     *
     * @param xml a chunked string
     * @param tag to be extracted
     *
     * @return true if the tag is found, false otherwise
     */
    private boolean getXmlFlag(ChunkedString xml, String tag) throws XmlException {
        ChunkedString val = null;

        if(XmlUtil.getTag(xml, tag) > -1) {
            val = XmlUtil.getTagValue(xml, tag);
            if(val.equals("true")) {
                return true;
            }
        }
        return false;
    }

    /**
     * Utility method to add a boolean xml flag into a String buffer
     * @param out the xml output stream
     * @param tag tag name
     * @param flag tag value (true, false)
     */
    private void addXmlFlag(StringBuffer out, String tag, boolean flag) {
        String val = flag ? "true" : "false" ;
        XmlUtil.addElement(out, tag, val);
    }

    /**
     * Utility method to get an xml date out of a chunked string
     * @param xml is the chunked string
     * @param tag the date tag
     * @return the date tag value
     */
    private Date getXmlDate(ChunkedString xml, String tag) throws XmlException {
        String val = null;
        MailDateFormatter df = new MailDateFormatter();

        if(XmlUtil.getTag(xml, tag) > -1) {
            val = XmlUtil.getTagValue(xml, tag).toString();
            return MailDateFormatter.parseUTCDate(val);
        }
        return null;
    }

    /**
     * Convert Funambol 3.0 folder names into local folder names.
     *
     * @param parent the item parent string
     * @return the converted name, or the original one if does not
     *         starts with "ROOT/"
     */
    private String folderFromParent(String parent) {
        final String root = "ROOT/";
        
        if(parent == null) {
            return null;
        }
        if (parent.startsWith(root)) {
            return folderFromPrefix(parent.charAt(root.length()));
        }
        return parent;
    }

    /**
     * Find the folder name from the Funambol key prefix.
     */
    private String folderFromPrefix(char prefix) {
        switch (prefix) {
            case 'I':
                return "Inbox";
            case 'O':
                return "Outbox";
            case 'D':
                return "Drafts";
            case 'S':
                return "Sent";
            default:
                return null;
        }
    }

    /**
     * This method creates a LUID for a given message.
     * In our client a LUID is folderName/messageRecordId
     *
     * @param msg the message
     * @return the LUID
     */
    private String makeLuid(Message msg) {
        String luid = msg.getParent().getName()+"/"+msg.getRecordId();
        return luid;
    }

    /**
     * Utility method to dump the message flags into a string representing a
     * SyncML email item.
     * @param flags message flags
     * @param out output string
     */
    private void formatFlags(MessageFlags flags, StringBuffer out) {
        addXmlFlag(out, READ_TAG, flags.isSet(MessageFlags.OPENED));
        addXmlFlag(out, FORW_TAG, flags.isSet(MessageFlags.FORWARDED));
        addXmlFlag(out, REPL_TAG, flags.isSet(MessageFlags.ANSWERED));
        addXmlFlag(out, DELE_TAG, flags.isSet(MessageFlags.DELETED));
        addXmlFlag(out, FLAG_TAG, flags.isSet(MessageFlags.FLAGGED));
    }

    /**
     * Compute root folders. This method initializes the rootFolders array with
     * the set of root folders. If the array has already been computed then the
     * method will overwrite it.
     */
    private void computeRootFolders() {

        if (rootFolders == null) {
            // Build the ordered root folder list
            Folder[] flist = msgStore.list();
            int len=flist.length;
            rootFolders = new Folder[len];
            
            for(int i=0, j=4; i<flist.length; i++) {
                String name = flist[i].getName();
                int idx = 0;
                
                if(name.equals(Store.INBOX)) {
                    idx = INBOX;
                } else if (name.equals(Store.OUTBOX)) {
                    idx = OUTBOX;
                } else if (name.equals(Store.DRAFTS)) {
                    idx = DRAFTS;
                } else if (name.equals(Store.SENT)) {
                    idx = SENT;
                } else {
                    idx = j++;
                }
               
                rootFolders[idx] = new Folder(flist[i].getFullName(), msgStore);
            }
        }
    }

}

